"""
Module to implement a plugin that ensures that the indentation for List Items
are equivalent with each other.
"""

import copy
from dataclasses import dataclass
from enum import Enum
from typing import Any, Dict, List, Optional, cast

from pymarkdown.general.parser_helper import ParserHelper
from pymarkdown.plugin_manager.plugin_details import PluginDetailsV2
from pymarkdown.plugin_manager.plugin_scan_context import PluginScanContext
from pymarkdown.plugin_manager.rule_plugin import RulePlugin
from pymarkdown.tokens.inline_code_span_markdown_token import (
    InlineCodeSpanMarkdownToken,
)
from pymarkdown.tokens.list_start_markdown_token import ListStartMarkdownToken
from pymarkdown.tokens.markdown_token import EndMarkdownToken, MarkdownToken
from pymarkdown.tokens.new_list_item_markdown_token import NewListItemMarkdownToken
from pymarkdown.tokens.ordered_list_start_markdown_token import (
    OrderedListStartMarkdownToken,
)
from pymarkdown.tokens.raw_html_markdown_token import RawHtmlMarkdownToken
from pymarkdown.tokens.reference_markdown_token import ReferenceMarkdownToken
from pymarkdown.tokens.text_markdown_token import TextMarkdownToken
from pymarkdown.tokens.unordered_list_start_markdown_token import (
    UnorderedListStartMarkdownToken,
)


class OrderedListAlignment(Enum):
    """
    Enumeration to provide guidance on what alignment was used for ordered lists.
    """

    UNKNOWN = 0
    LEFT = 1
    RIGHT = 2


@dataclass
class LeadingSpaceAdjustment:
    """
    Internal class to represent an adjustment to be made.
    """

    token: Any
    start_index: int
    end_index: int
    delta_indent: int


@dataclass
class DeferredAdjustmant:
    """
    Internal class to hold any deferred adjustments to the leading spaces.
    """

    token: MarkdownToken
    line_count: int
    delta_indent: int


# pylint: disable=too-many-instance-attributes,protected-access
class RuleMd005(RulePlugin):
    """
    Class to implement a plugin that ensures that the indentation for List Items
    are equivalent with each other.
    """

    def __init__(self) -> None:
        super().__init__()
        self.__list_stack: List[ListStartMarkdownToken] = []
        self.__read_only_list_stack: List[ListStartMarkdownToken] = []

        self.__leading_space_adjustments: Dict[
            MarkdownToken, List[LeadingSpaceAdjustment]
        ] = {}
        self.__unordered_list_indents: Dict[int, int] = {}
        self.__unordered_current_indents: Dict[int, int] = {}
        self.__unordered_did_first: Dict[int, bool] = {}
        self.__ordered_list_starts: Dict[int, OrderedListStartMarkdownToken] = {}
        self.__ordered_tokens: Dict[int, List[MarkdownToken]] = {}
        self.__ordered_list_alignment: Dict[int, OrderedListAlignment] = {}
        self.__line_count: Dict[int, int] = {}
        self.__deferred_adjustment: Dict[int, DeferredAdjustmant] = {}
        self.__list_start_indices: Dict[int, Dict[MarkdownToken, int]] = {}
        self.__list_end_indices: Dict[int, Dict[MarkdownToken, int]] = {}
        self.__current_list_tokens: Dict[int, MarkdownToken] = {}
        # self.__debug = False

    def get_details(self) -> PluginDetailsV2:
        """
        Get the details for the plugin.
        """
        return PluginDetailsV2(
            plugin_name="list-indent",
            plugin_id="MD005",
            plugin_enabled_by_default=True,
            plugin_description="Inconsistent indentation for list items at the same level",
            plugin_version="0.5.2",
            plugin_url="https://pymarkdown.readthedocs.io/en/latest/plugins/rule_md005.md",
            plugin_supports_fix=True,
            plugin_fix_level=2,
        )

    def starting_new_file(self) -> None:
        """
        Event that the a new file to be scanned is starting.
        """
        self.__list_stack = []
        self.__read_only_list_stack = []
        self.__leading_space_adjustments = {}
        self.__unordered_list_indents = {}
        self.__unordered_current_indents = {}
        self.__unordered_did_first = {}
        self.__ordered_list_starts = {}
        self.__ordered_tokens = {}
        self.__ordered_list_alignment = {}
        self.__line_count = {}
        self.__deferred_adjustment = {}
        self.__list_start_indices = {}
        self.__list_end_indices = {}
        self.__current_list_tokens = {}

    def __report_issue_new_list_item(
        self,
        context: PluginScanContext,
        token: MarkdownToken,
        list_level: int,
        delta: Optional[int],
    ) -> None:
        list_token = cast(NewListItemMarkdownToken, token)
        base_token = self.__list_stack[-1]
        if base_token.is_unordered_list_start:
            expected_indent = self.__unordered_list_indents[list_level]
            expected_indent = self.__unordered_current_indents[list_level]
            delta_indent = list_token.indent_level - expected_indent
            # if delta is not None:
            #     delta_indent += delta
        else:
            expected_indent = base_token.indent_level - len(
                base_token.list_start_content
            )
            this_indent = list_token.indent_level - len(list_token.list_start_content)
            delta_indent = this_indent - expected_indent
            if delta and delta < 0:
                delta_indent -= delta

            expected_indent = list_token.indent_level - delta_indent

        adjusted_whitespace = (
            list_token.extracted_whitespace[:-delta_indent]
            if delta_indent > 0
            else list_token.extracted_whitespace + " " * (-delta_indent)
        )
        self.register_fix_token_request(
            context, token, "next_token", "indent_level", expected_indent
        )
        self.register_fix_token_request(
            context,
            token,
            "next_token",
            "extracted_whitespace",
            adjusted_whitespace,
        )
        if base_token.leading_spaces:
            self.__deferred_adjustment[list_level] = DeferredAdjustmant(
                token,
                self.__line_count[list_level],
                delta_indent,
            )

    def __report_issue_ordered_list(
        self, context: PluginScanContext, token: MarkdownToken, list_level: int
    ) -> None:
        list_copy_token = cast(OrderedListStartMarkdownToken, self.__list_stack[-1])
        list_token = cast(OrderedListStartMarkdownToken, token)
        base_token = self.__ordered_list_starts[list_level]
        expected_indent = base_token.column_number
        delta_indent = list_token.column_number - expected_indent
        adjusted_whitespace = (
            list_token.extracted_whitespace[:-delta_indent]
            if delta_indent > 0
            else list_token.extracted_whitespace + (" " * (-delta_indent))
        )
        self.register_fix_token_request(
            context,
            token,
            "next_token",
            "indent_level",
            list_token.indent_level - delta_indent,
        )
        list_copy_token._modify_token(
            "indent_level", list_token.indent_level - delta_indent
        )
        self.register_fix_token_request(
            context,
            token,
            "next_token",
            "extracted_whitespace",
            adjusted_whitespace,
        )
        list_copy_token._modify_token("extracted_whitespace", adjusted_whitespace)
        self.register_fix_token_request(
            context, token, "next_token", "column_number", expected_indent
        )
        list_copy_token._modify_token("column_number", expected_indent)
        if list_token.leading_spaces:
            self.__deferred_adjustment[list_level] = DeferredAdjustmant(
                token, self.__line_count[list_level], delta_indent
            )

    def __report_issue_unordered_list(
        self, context: PluginScanContext, token: MarkdownToken, list_level: int
    ) -> None:
        list_token = cast(UnorderedListStartMarkdownToken, token)
        expected_indent = self.__unordered_list_indents[list_level]
        delta_indent = list_token.indent_level - expected_indent

        previous_list_level = list_level - 1
        if previous_list_level in self.__ordered_tokens:
            previous_ordered_token = cast(
                OrderedListStartMarkdownToken,
                self.__ordered_tokens[previous_list_level][-1],
            )
            start_content_length = len(previous_ordered_token.list_start_content)
            previous_token = self.__ordered_list_starts[previous_list_level]
            indent_adjust = start_content_length - len(
                previous_token.list_start_content
            )
            delta_indent += indent_adjust
            expected_indent -= indent_adjust

            self.__unordered_current_indents[list_level] = (
                self.__unordered_current_indents[list_level] - indent_adjust
            )
        else:
            self.__unordered_current_indents[list_level] = expected_indent

        adjusted_whitespace = (
            list_token.extracted_whitespace[:-delta_indent]
            if delta_indent > 0
            else list_token.extracted_whitespace + (" " * (-delta_indent))
        )
        self.register_fix_token_request(
            context, token, "next_token", "indent_level", expected_indent
        )
        self.register_fix_token_request(
            context,
            token,
            "next_token",
            "extracted_whitespace",
            adjusted_whitespace,
        )
        self.register_fix_token_request(
            context,
            token,
            "next_token",
            "column_number",
            list_token.column_number - delta_indent,
        )
        if list_token.leading_spaces:
            self.__deferred_adjustment[list_level] = DeferredAdjustmant(
                token,
                self.__line_count[list_level],
                delta_indent,
            )

    def __report_issue(
        self,
        context: PluginScanContext,
        token: MarkdownToken,
        expected_indent: int,
        delta: Optional[int] = None,
    ) -> None:
        if context.in_fix_mode:
            list_level = len(self.__list_stack)
            if token.is_new_list_item:
                self.__report_issue_new_list_item(context, token, list_level, delta)
            elif token.is_unordered_list_start:
                self.__report_issue_unordered_list(context, token, list_level)
            else:
                self.__report_issue_ordered_list(context, token, list_level)
        else:
            if expected_indent < 0:
                list_token = cast(NewListItemMarkdownToken, token)
                list_level = len(self.__list_stack)
                expected_indent = (
                    self.__ordered_list_starts[list_level].indent_level
                    - 2
                    - len(list_token.list_start_content)
                )

            extra_data = (
                f"Expected: {expected_indent}; Actual: {token.column_number - 1}"
            )
            # if self.__debug:
            #     print(f"ERROR>>{extra_data}")
            self.report_next_token_error(context, token, extra_data)

    def __handle_ordered_list_item_right(
        self,
        list_level: int,
        context: PluginScanContext,
        token: OrderedListStartMarkdownToken,
    ) -> None:
        assert self.__ordered_list_starts[list_level].extracted_whitespace
        original_text = (
            self.__ordered_list_starts[list_level].list_start_content
            + self.__ordered_list_starts[list_level].extracted_whitespace
        )
        prefix_text = f"{token.list_start_content}{token.extracted_whitespace}"
        original_text_length = len(original_text)
        current_prefix_length = len(prefix_text)
        if delta := original_text_length - current_prefix_length:
            # if self.__debug:
            #     print("ri1")
            self.__report_issue(context, token, -1, delta)
        else:
            assert (
                token.indent_level
                == self.__ordered_list_starts[list_level].indent_level
            )

    def __handle_ordered_list_item(
        self, context: PluginScanContext, token: OrderedListStartMarkdownToken
    ) -> None:
        list_level = len(self.__list_stack)
        list_alignment = self.__ordered_list_alignment[list_level]

        self.__handle_leading_space_adjustments(context, token, False)

        if list_alignment == OrderedListAlignment.RIGHT:
            self.__handle_ordered_list_item_right(list_level, context, token)
        elif (
            self.__ordered_list_starts[list_level].column_number != token.column_number
        ):
            # if self.__debug:
            #     print("ri2")
            self.__report_issue(
                context, token, self.__ordered_list_starts[list_level].column_number - 1
            )

    def __compute_ordered_list_alignment(self) -> None:
        list_level = len(self.__list_stack)

        last_length = 0
        last_token: Optional[NewListItemMarkdownToken] = None

        for next_token in self.__ordered_tokens[list_level]:
            list_token = cast(NewListItemMarkdownToken, next_token)
            content_length = len(list_token.list_start_content)
            if not last_length:
                last_length = content_length
                last_token = list_token
            elif content_length != last_length:
                assert last_token is not None
                if last_token.column_number == list_token.column_number:
                    self.__ordered_list_alignment[list_level] = (
                        OrderedListAlignment.LEFT
                    )
                    break
                last_total_length = len(last_token.extracted_whitespace) + len(
                    last_token.list_start_content
                )
                next_total_length = len(list_token.extracted_whitespace) + len(
                    list_token.list_start_content
                )
                if last_total_length == next_total_length:
                    self.__ordered_list_alignment[list_level] = (
                        OrderedListAlignment.RIGHT
                    )
                    break

    def __calculate_indent_level(self, list_level: int) -> int:

        list_indent = self.__unordered_list_indents[list_level]
        did_first = self.__unordered_did_first[list_level]
        indent_level = list_level
        while not did_first and indent_level > 0:
            if self.__list_stack[indent_level - 1].is_ordered_list_start:
                ordered_list_items = self.__ordered_tokens[indent_level]
                if len(ordered_list_items) > 1:
                    first_item = cast(
                        OrderedListStartMarkdownToken, ordered_list_items[0]
                    )
                    delta = len(
                        cast(
                            OrderedListStartMarkdownToken, ordered_list_items[-1]
                        ).list_start_content
                    ) - len(first_item.list_start_content)
                    list_indent += delta
                break
            indent_level -= 1
        return list_indent

    def __handle_unordered_list_start(
        self, context: PluginScanContext, token: UnorderedListStartMarkdownToken
    ) -> None:
        self.__list_stack.append(token)
        self.__read_only_list_stack.append(token)
        list_level = len(self.__list_stack)
        self.__line_count[list_level] = 0

        did_first = False
        if list_level not in self.__unordered_list_indents:
            self.__unordered_list_indents[list_level] = token.indent_level
            did_first = True
        self.__unordered_current_indents[list_level] = token.indent_level
        self.__unordered_did_first[list_level] = did_first

        if self.__calculate_indent_level(list_level) != token.indent_level:
            # if self.__debug:
            #     print("ri3")
            first_indent = self.__unordered_list_indents[list_level] - 2
            second_indent = token.column_number - 1
            if first_indent != second_indent:
                self.__report_issue(
                    context, token, self.__unordered_list_indents[list_level] - 2
                )

    def __handle_ordered_list_start(self, token: OrderedListStartMarkdownToken) -> None:
        token_copy = copy.deepcopy(token)
        self.__list_stack.append(token_copy)
        self.__read_only_list_stack.append(token)
        list_level = len(self.__list_stack)
        self.__line_count[list_level] = 0

        new_map: Dict[MarkdownToken, int] = {}
        self.__list_start_indices[list_level] = new_map
        new_map[token] = self.__line_count[list_level]

        new_map = {}
        self.__list_end_indices[list_level] = new_map
        self.__current_list_tokens[list_level] = token

        self.__ordered_tokens[list_level] = [token]
        if list_level not in self.__ordered_list_starts:
            self.__ordered_list_starts[list_level] = token_copy
            self.__ordered_list_alignment[list_level] = OrderedListAlignment.UNKNOWN

    def __handle_deferred_adjustment(self) -> None:
        list_stack_length = len(self.__list_stack)
        current_deferred = self.__deferred_adjustment[list_stack_length]
        token = current_deferred.token
        delta_indent = current_deferred.delta_indent

        if list_stack_length in self.__list_start_indices:
            start_index = self.__list_start_indices[list_stack_length][token]
            end_index = self.__list_end_indices[list_stack_length][token]
        else:
            start_index = current_deferred.line_count
            end_index = self.__line_count[list_stack_length]

        owner = self.__read_only_list_stack[-1]
        owner_list: List[LeadingSpaceAdjustment] = []
        if owner in self.__leading_space_adjustments:
            owner_list = self.__leading_space_adjustments[owner]
        else:
            self.__leading_space_adjustments[owner] = owner_list

        owner_tuple = LeadingSpaceAdjustment(
            token, start_index, end_index, delta_indent
        )
        owner_list.append(owner_tuple)

    def __handle_leading_space_adjustments(
        self,
        context: PluginScanContext,
        token: MarkdownToken,
        do_leading_adjustments: bool,
    ) -> None:
        list_stack_length = len(self.__list_stack)
        if list_stack_length in self.__deferred_adjustment:
            self.__handle_deferred_adjustment()
            del self.__deferred_adjustment[list_stack_length]

        owner = self.__read_only_list_stack[-1]
        if do_leading_adjustments and owner in self.__leading_space_adjustments:
            end_token = cast(EndMarkdownToken, token)
            list_start_token = cast(
                ListStartMarkdownToken, end_token.start_markdown_token
            )
            assert list_start_token.leading_spaces is not None
            split_leading_spaces = list_start_token.leading_spaces.split("\n")
            owner_list = self.__leading_space_adjustments[owner]
            for next_adjustment in owner_list:
                for i in range(next_adjustment.start_index, next_adjustment.end_index):
                    split_leading_spaces[i] = (
                        split_leading_spaces[i][: -next_adjustment.delta_indent]
                        if next_adjustment.delta_indent > 0
                        else split_leading_spaces[i]
                        + (" " * -next_adjustment.delta_indent)
                    )
            self.register_fix_token_request(
                context,
                end_token.start_markdown_token,
                "next_token",
                "leading_spaces",
                "\n".join(split_leading_spaces),
            )

    def __handle_list_item(
        self, context: PluginScanContext, token: NewListItemMarkdownToken
    ) -> None:
        list_level = len(self.__list_stack)
        if list_level in self.__list_start_indices:
            current_list_token = self.__current_list_tokens[list_level]
            self.__current_list_tokens[list_level] = token

            self.__list_end_indices[list_level][current_list_token] = self.__line_count[
                list_level
            ]
            self.__list_start_indices[list_level][token] = self.__line_count[list_level]

        self.__handle_leading_space_adjustments(context, token, False)

        if self.__list_stack[-1].is_unordered_list_start:
            list_level = len(self.__list_stack)
            xx = self.__calculate_indent_level(list_level)
            if xx != token.indent_level:
                delta_indent = token.indent_level - xx if context.in_fix_mode else None
                self.__report_issue(
                    context,
                    token,
                    self.__unordered_list_indents[len(self.__list_stack)] - 2,
                    delta_indent,
                )
        else:
            self.__ordered_tokens[len(self.__list_stack)].append(token)

    def __handle_list_end(
        self, context: PluginScanContext, token: EndMarkdownToken
    ) -> None:
        list_level = len(self.__list_stack)
        if list_level in self.__list_start_indices:
            current_list_token = self.__current_list_tokens[list_level]
            self.__list_end_indices[list_level][current_list_token] = self.__line_count[
                list_level
            ]

        self.__handle_leading_space_adjustments(context, token, False)

        if token.is_ordered_list_end:
            list_level = len(self.__list_stack)
            if (
                self.__ordered_list_alignment[list_level]
                == OrderedListAlignment.UNKNOWN
            ):
                self.__compute_ordered_list_alignment()
            for next_token in self.__ordered_tokens[list_level]:
                next_list_token = cast(OrderedListStartMarkdownToken, next_token)
                self.__handle_ordered_list_item(context, next_list_token)
        self.__handle_leading_space_adjustments(context, token, True)
        del self.__line_count[len(self.__list_stack)]
        del self.__read_only_list_stack[-1]
        del self.__list_stack[-1]
        if list_level in self.__list_start_indices:
            del self.__list_start_indices[list_level]
            del self.__list_end_indices[list_level]
            del self.__current_list_tokens[list_level]
        if not self.__list_stack:
            self.__unordered_list_indents = {}
            self.__unordered_current_indents = {}
            self.__ordered_list_starts = {}
            self.__unordered_did_first = {}

    def __count_newlines_in_token(self, current_token: MarkdownToken) -> int:
        newlines_in_text_token = 0
        if current_token.is_text:
            text_token = cast(TextMarkdownToken, current_token)
            newlines_in_text_token = ParserHelper.count_newlines_in_text(
                text_token.token_text
            )
        elif current_token.is_inline_image or current_token.is_inline_link:
            reference_token = cast(ReferenceMarkdownToken, current_token)
            newlines_in_text_token += ParserHelper.count_newlines_in_text(
                reference_token.text_from_blocks
            )
        elif current_token.is_inline_raw_html:
            raw_html_token = cast(RawHtmlMarkdownToken, current_token)
            newlines_in_text_token += ParserHelper.count_newlines_in_text(
                raw_html_token.raw_tag
            )
        elif current_token.is_inline_code_span:
            code_span_token = cast(InlineCodeSpanMarkdownToken, current_token)
            newlines_in_text_token += ParserHelper.count_newlines_in_texts(
                code_span_token.leading_whitespace,
                code_span_token.span_text,
                code_span_token.trailing_whitespace,
            )
        return newlines_in_text_token

    def next_token(self, context: PluginScanContext, token: MarkdownToken) -> None:
        """
        Event that a new token is being processed.
        """
        # print(f"token>>{ParserHelper.make_value_visible(token)}")
        if token.is_unordered_list_start:
            unordered_list_start_token = cast(UnorderedListStartMarkdownToken, token)
            self.__handle_unordered_list_start(context, unordered_list_start_token)
        elif token.is_ordered_list_start:
            ordered_list_start_token = cast(OrderedListStartMarkdownToken, token)
            self.__handle_ordered_list_start(ordered_list_start_token)
        elif token.is_unordered_list_end or token.is_ordered_list_end:
            end_token = cast(EndMarkdownToken, token)
            self.__handle_list_end(context, end_token)
        elif token.is_new_list_item:
            new_list_item_token = cast(NewListItemMarkdownToken, token)
            self.__handle_list_item(context, new_list_item_token)
        elif list_stack_length := len(self.__list_stack):
            self.__line_count[list_stack_length] += self.__count_newlines_in_token(
                token
            )


# pylint: enable=too-many-instance-attributes,protected-access



"""
Module to provide for an inline element that can be added to markdown parsing stream.
"""

import logging
from typing import Optional

from pymarkdown.general.parser_logger import ParserLogger
from pymarkdown.general.position_marker import PositionMarker
from pymarkdown.tokens.markdown_token import MarkdownToken, MarkdownTokenClass

POGGER = ParserLogger(logging.getLogger(__name__))


# pylint: disable=too-many-arguments
class InlineMarkdownToken(MarkdownToken):
    """
    Class to provide for a leaf element that can be added to markdown parsing stream.
    """

    def __init__(
        self,
        token_name: str,
        extra_data: Optional[str],
        line_number: int = 0,
        column_number: int = 0,
        position_marker: Optional[PositionMarker] = None,
        requires_end_token: bool = False,
        can_force_close: bool = True,
        is_special: bool = False,
    ):
        MarkdownToken.__init__(
            self,
            token_name,
            MarkdownTokenClass.INLINE_BLOCK,
            extra_data,
            line_number=line_number,
            column_number=column_number,
            position_marker=position_marker,
            can_force_close=can_force_close,
            requires_end_token=requires_end_token,
            is_special=is_special,
        )


# pylint: enable=too-many-arguments



"""
Module to implement a plugin that ensures that specific proper names have
the correct capitalization.
"""

from dataclasses import dataclass
from typing import List, Optional, cast

from pymarkdown.general.constants import Constants
from pymarkdown.general.parser_helper import ParserHelper
from pymarkdown.plugin_manager.plugin_details import (
    PluginDetailsV2,
    PluginDetailsV3,
    QueryConfigItem,
)
from pymarkdown.plugin_manager.plugin_scan_context import PluginScanContext
from pymarkdown.plugin_manager.rule_plugin import RulePlugin
from pymarkdown.tokens.image_start_markdown_token import ImageStartMarkdownToken
from pymarkdown.tokens.inline_code_span_markdown_token import (
    InlineCodeSpanMarkdownToken,
)
from pymarkdown.tokens.link_reference_definition_markdown_token import (
    LinkReferenceDefinitionMarkdownToken,
)
from pymarkdown.tokens.link_start_markdown_token import LinkStartMarkdownToken
from pymarkdown.tokens.markdown_token import EndMarkdownToken, MarkdownToken
from pymarkdown.tokens.text_markdown_token import TextMarkdownToken


@dataclass
class FoundReplacement:
    """
    Class containing replacements found.
    """

    found_index: int
    required_capitalization: str
    part_context: str


class RuleMd044(RulePlugin):
    """
    Class to implement a plugin that ensures that specific proper names have
    the correct capitalization.
    """

    def __init__(self) -> None:
        super().__init__()
        self.__proper_name_list: List[str] = []
        self.__check_in_code_blocks: bool = False
        self.__check_in_code_spans: bool = False
        self.__is_in_code_block: bool = False
        self.__names: str = ""
        self.__replacement_items: List[FoundReplacement] = []

    def get_details(self) -> PluginDetailsV2:
        """
        Get the details for the plugin.
        """
        return PluginDetailsV3(
            plugin_name="proper-names",
            plugin_id="MD044",
            plugin_enabled_by_default=True,
            plugin_description="Proper names should have the correct capitalization",
            plugin_version="0.7.1",
            plugin_url="https://pymarkdown.readthedocs.io/en/latest/plugins/rule_md044.md",
            plugin_configuration="names,code_blocks",
            plugin_supports_fix=True,
            plugin_fix_level=2,
        )

    def starting_new_file(self) -> None:
        """
        Event that the a new file to be scanned is starting.
        """
        self.__is_in_code_block = False

    def initialize_from_config(self) -> None:
        """
        Event to allow the plugin to load configuration information.
        """
        self.__check_in_code_blocks = (
            self.plugin_configuration.get_boolean_property_with_default(
                "code_blocks", True
            )
        )
        self.__check_in_code_spans = (
            self.plugin_configuration.get_boolean_property_with_default(
                "code_spans", True
            )
        )
        self.__proper_name_list = []
        self.__names = self.plugin_configuration.get_string_property_with_default(
            "names",
            "",
        ).strip(" ")
        if self.__names:
            lower_list: List[str] = []
            for next_name in self.__names.split(","):
                next_name = next_name.strip(" ")
                if not next_name:
                    raise ValueError(
                        "Elements in the comma-separated list cannot be empty."
                    )
                if next_name.lower() in lower_list:
                    raise ValueError(
                        f"Element `{next_name}` is already present in the list as "
                        + f"`{self.__proper_name_list[lower_list.index(next_name.lower())]}`."
                    )
                lower_list.append(next_name.lower())
                self.__proper_name_list.append(next_name)

    def query_config(self) -> List[QueryConfigItem]:
        """
        Query to find out the configuration that the rule is using.
        """
        return [
            QueryConfigItem("code_blocks", self.__check_in_code_blocks),
            QueryConfigItem("code_spans", self.__check_in_code_spans),
            QueryConfigItem("names", self.__names),
        ]

    # pylint: disable=too-many-arguments
    def __check_for_proper_match(
        self,
        original_source: str,
        found_index: int,
        required_capitalization: str,
        context: PluginScanContext,
        token: MarkdownToken,
        line_adjust: int,
        col_adjust: int,
        part_context: str,
    ) -> None:
        original_found_text = original_source[
            found_index : found_index + len(required_capitalization)
        ]
        after_found_index = found_index + len(required_capitalization)

        is_character_before_match = (
            original_source[found_index - 1].isalnum() if found_index > 0 else False
        )
        is_character_after_match = (
            original_source[after_found_index].isalnum()
            if after_found_index < len(original_source)
            else False
        )
        if (
            not is_character_after_match
            and not is_character_before_match
            and original_found_text != required_capitalization
        ):
            assert len(original_found_text) == len(required_capitalization)
            if context.in_fix_mode:
                # d1 = max(0,found_index-5)
                # d2 = min(len(original_source), found_index+5)
                # dd  = original_source[d1:d2]
                self.__replacement_items.append(
                    FoundReplacement(found_index, required_capitalization, part_context)
                )
            else:
                extra_data = f"Expected: {required_capitalization}; Actual: {original_found_text}"
                self.report_next_token_error(
                    context,
                    token,
                    extra_error_information=extra_data,
                    line_number_delta=line_adjust
                    + context.calc_pragma_offset(token, line_adjust),
                    column_number_delta=col_adjust,
                )

    # pylint: enable=too-many-arguments

    # pylint: disable=too-many-arguments
    def __search_for_possible_matches(
        self,
        string_to_check: str,
        string_to_check_lower: str,
        search_start: int,
        found_index: int,
        start_x_offset: int,
        start_y_offset: int,
        same_line_offset: int,
        next_name: str,
        context: PluginScanContext,
        token: MarkdownToken,
        part_context: str,
    ) -> None:
        col_adjust, line_adjust = ParserHelper.adjust_for_newlines(
            string_to_check_lower, search_start, found_index
        )
        if line_adjust == 0 and start_y_offset == 0:
            assert col_adjust >= 0
            col_adjust -= same_line_offset
            assert same_line_offset <= 0
        line_adjust += start_y_offset
        if col_adjust == 0 and start_x_offset:
            col_adjust += (
                -start_x_offset if start_x_offset > 0 else -(-start_x_offset - 1)
            )
            col_adjust = -col_adjust
        elif col_adjust > 0 and start_x_offset:
            col_adjust += -start_x_offset - 1
            col_adjust = -col_adjust
        self.__check_for_proper_match(
            string_to_check,
            found_index,
            next_name,
            context,
            token,
            line_adjust,
            col_adjust,
            part_context,
        )

    # pylint: enable=too-many-arguments

    # pylint: disable=too-many-arguments
    def __search_for_matches(
        self,
        string_to_check: str,
        context: PluginScanContext,
        token: MarkdownToken,
        same_line_offset: int = 0,
        start_x_offset: int = 0,
        start_y_offset: int = 0,
        part_context: str = "",
        keep_text_with_markers: bool = False,
    ) -> None:
        if not keep_text_with_markers:
            string_to_check = ParserHelper.remove_all_from_text(string_to_check)
        string_to_check_lower = string_to_check.lower()
        for next_name in self.__proper_name_list:
            next_name_lower = next_name.lower()
            search_start = 0
            found_index = string_to_check_lower.find(next_name_lower, search_start)
            while found_index != -1:
                self.__search_for_possible_matches(
                    string_to_check,
                    string_to_check_lower,
                    search_start,
                    found_index,
                    start_x_offset,
                    start_y_offset,
                    same_line_offset,
                    next_name,
                    context,
                    token,
                    part_context,
                )

                search_start = found_index + len(next_name)
                found_index = string_to_check_lower.find(next_name_lower, search_start)

    # pylint: enable=too-many-arguments

    # pylint: disable=too-many-arguments
    def __adjust_for_newlines_and_search(
        self,
        context: PluginScanContext,
        token: MarkdownToken,
        link_body_text: str,
        full_link_text: str,
        string_to_check: str,
        same_line_offset: int,
        part_context: str,
    ) -> None:
        start_x_offset = 0
        start_y_offset = 0
        if ParserHelper.newline_character in link_body_text:
            start_x_offset, start_y_offset = ParserHelper.adjust_for_newlines(
                full_link_text, 0, len(full_link_text)
            )
        self.__search_for_matches(
            string_to_check,
            context,
            token,
            same_line_offset,
            start_x_offset,
            start_y_offset,
            part_context,
        )

    # pylint: enable=too-many-arguments

    def __handle_inline_link_end(
        self, context: PluginScanContext, token: MarkdownToken
    ) -> None:
        end_token = cast(EndMarkdownToken, token)
        link_token = cast(LinkStartMarkdownToken, end_token.start_markdown_token)
        if link_token.label_type == Constants.link_type__inline:
            assert link_token.before_title_whitespace is not None
            assert link_token.inline_title_bounding_character is not None
            assert link_token.before_link_whitespace is not None
            link_body = "".join(
                [
                    link_token.before_link_whitespace,
                    link_token.active_link_uri,
                    link_token.before_title_whitespace,
                    link_token.inline_title_bounding_character,
                ]
            )
            full_link_text = "".join(
                [
                    "[",
                    link_token.text_from_blocks,
                    "](",
                    link_body,
                ]
            )
            same_line_offset = -(len(full_link_text))
            assert link_token.active_link_title is not None
            self.__adjust_for_newlines_and_search(
                context,
                link_token,
                link_body,
                full_link_text,
                link_token.active_link_title,
                same_line_offset,
                "x",
            )

    def __handle_inline_link_fix(
        self, context: PluginScanContext, token: MarkdownToken
    ) -> None:
        link_token = cast(LinkStartMarkdownToken, token)
        assert link_token.link_title is not None
        self.__search_for_matches(
            link_token.link_title, context, token, 0, part_context="link_title"
        )
        self.__search_for_matches(
            link_token.text_from_blocks,
            context,
            token,
            0,
            part_context="text_from_blocks",
        )
        if (
            link_token.label_type == Constants.link_type__inline
            and link_token.pre_link_title
        ):
            self.__search_for_matches(
                link_token.pre_link_title,
                context,
                token,
                0,
                part_context="pre_link_title",
            )

    def __handle_inline_image(
        self, context: PluginScanContext, token: MarkdownToken
    ) -> None:  # sourcery skip: extract-method
        link_token = cast(LinkStartMarkdownToken, token)
        same_line_offset = -2
        self.__search_for_matches(
            link_token.text_from_blocks, context, token, same_line_offset
        )

        if (
            link_token.label_type == Constants.link_type__inline
            and not context.in_fix_mode
        ):
            assert link_token.before_link_whitespace is not None
            assert link_token.before_title_whitespace is not None
            assert link_token.inline_title_bounding_character is not None
            link_body = "".join(
                [
                    link_token.before_link_whitespace,
                    link_token.active_link_uri,
                    link_token.before_title_whitespace,
                    link_token.inline_title_bounding_character,
                ]
            )
            full_link_text = f"![{link_token.text_from_blocks}]({link_body}"
            same_line_offset = -(len(full_link_text))
            assert link_token.active_link_title is not None
            self.__adjust_for_newlines_and_search(
                context,
                token,
                link_body,
                full_link_text,
                link_token.active_link_title,
                same_line_offset,
                "y",
            )
        elif context.in_fix_mode:
            image_token = cast(ImageStartMarkdownToken, token)
            assert image_token.link_title is not None
            self.__search_for_matches(
                image_token.link_title, context, token, 0, part_context="link_title"
            )
            self.__search_for_matches(
                image_token.text_from_blocks,
                context,
                token,
                0,
                part_context="text_from_blocks",
            )

    def __handle_link_reference_definition(
        self, context: PluginScanContext, token: MarkdownToken
    ) -> None:
        lrd_token = cast(LinkReferenceDefinitionMarkdownToken, token)
        link_name = lrd_token.link_name_debug or lrd_token.link_name
        same_line_offset = -1
        if context.in_fix_mode:
            if lrd_token.link_name_debug:
                self.__search_for_matches(
                    lrd_token.link_name_debug,
                    context,
                    token,
                    same_line_offset,
                    part_context="link_name_debug",
                )
            self.__search_for_matches(
                lrd_token.link_name,
                context,
                token,
                same_line_offset,
                part_context="link_name",
            )
        else:
            self.__search_for_matches(
                link_name, context, token, same_line_offset, part_context="link_name"
            )

        assert lrd_token.link_destination_whitespace is not None
        assert lrd_token.link_destination is not None
        assert lrd_token.link_title_whitespace is not None
        full_link_text = "".join(
            [
                "[",
                link_name,
                "]:",
                lrd_token.link_destination_whitespace,
                lrd_token.link_destination,
                lrd_token.link_title_whitespace,
                "'",
            ]
        )
        same_line_offset = -(len(full_link_text) - 1)
        assert lrd_token.link_title_raw is not None
        if context.in_fix_mode:
            self.__adjust_for_newlines_and_search(
                context,
                token,
                full_link_text,
                full_link_text,
                lrd_token.link_title_raw,
                same_line_offset,
                "link_title_raw",
            )
            if lrd_token.link_title:
                self.__adjust_for_newlines_and_search(
                    context,
                    token,
                    full_link_text,
                    full_link_text,
                    lrd_token.link_title,
                    same_line_offset,
                    "link_title",
                )
        else:
            self.__adjust_for_newlines_and_search(
                context,
                token,
                full_link_text,
                full_link_text,
                lrd_token.link_title_raw,
                same_line_offset,
                "link_title_raw",
            )

    def __handle_inline_code_span(
        self, context: PluginScanContext, token: MarkdownToken
    ) -> None:
        if not self.__check_in_code_spans:
            return
        code_span_token = cast(InlineCodeSpanMarkdownToken, token)
        same_line_offset = -(
            len(code_span_token.extracted_start_backticks)
            + len(code_span_token.leading_whitespace)
        )
        self.__search_for_matches(
            code_span_token.span_text,
            context,
            token,
            same_line_offset,
            keep_text_with_markers=context.in_fix_mode,
        )

    def __handle_text(self, context: PluginScanContext, token: MarkdownToken) -> None:
        text_token = cast(TextMarkdownToken, token)
        if not self.__is_in_code_block or self.__check_in_code_blocks:
            self.__search_for_matches(
                text_token.token_text,
                context,
                token,
                keep_text_with_markers=context.in_fix_mode,
            )

    def next_token(self, context: PluginScanContext, token: MarkdownToken) -> None:
        """
        Event that a new token is being processed.
        """
        if not self.__proper_name_list:
            return
        self.__replacement_items.clear()
        if token.is_text:
            self.__handle_text(context, token)
        elif token.is_inline_code_span:
            self.__handle_inline_code_span(context, token)
        elif token.is_inline_link_end and not context.in_fix_mode:
            self.__handle_inline_link_end(context, token)
        elif token.is_inline_image:
            self.__handle_inline_image(context, token)
        elif token.is_link_reference_definition:
            self.__handle_link_reference_definition(context, token)
        elif token.is_code_block:
            self.__is_in_code_block = True
        elif token.is_code_block_end:
            self.__is_in_code_block = False
        elif context.in_fix_mode:
            if token.is_inline_link:
                self.__handle_inline_link_fix(context, token)

        if self.__replacement_items:
            self.__apply_replacement_items(context, token)

    def __apply_replacement_items(
        self, context: PluginScanContext, token: MarkdownToken
    ) -> None:
        if token.is_text:
            text_token = cast(TextMarkdownToken, token)
            self.__apply_normal_replacement(
                context, token, text_token.token_text, "token_text"
            )
        elif token.is_inline_code_span:
            code_span_token = cast(InlineCodeSpanMarkdownToken, token)
            self.__apply_normal_replacement(
                context, token, code_span_token.span_text, "span_text"
            )
        elif token.is_link_reference_definition:
            lrd_token = cast(LinkReferenceDefinitionMarkdownToken, token)
            self.__apply_matching_replacement(
                context, token, lrd_token.link_name, "link_name"
            )
            if lrd_token.link_name_debug:
                self.__apply_matching_replacement(
                    context, token, lrd_token.link_name_debug, "link_name_debug"
                )
            self.__apply_matching_replacement(
                context, token, lrd_token.link_title_raw, "link_title_raw"
            )
            if lrd_token.link_title:
                self.__apply_matching_replacement(
                    context, token, lrd_token.link_title, "link_title"
                )
        elif token.is_inline_link:
            link_token = cast(LinkStartMarkdownToken, token)
            self.__apply_matching_replacement(
                context, token, link_token.link_title, "link_title"
            )
            self.__apply_matching_replacement(
                context, token, link_token.pre_link_title, "pre_link_title"
            )
            self.__apply_matching_replacement(
                context, token, link_token.text_from_blocks, "text_from_blocks"
            )
        else:
            assert (
                token.is_inline_image
            ), "If not anything else, must be an inline image."
            image_token = cast(ImageStartMarkdownToken, token)
            self.__apply_matching_replacement(
                context, token, image_token.link_title, "link_title"
            )
            self.__apply_matching_replacement(
                context, token, image_token.text_from_blocks, "text_from_blocks"
            )

    def __apply_normal_replacement(
        self,
        context: PluginScanContext,
        token: MarkdownToken,
        text_to_check: Optional[str],
        name_to_check: str,
    ) -> None:
        assert text_to_check is not None
        new_text = text_to_check
        for replacement_item in self.__replacement_items:
            new_text = (
                new_text[: replacement_item.found_index]
                + replacement_item.required_capitalization
                + new_text[
                    replacement_item.found_index
                    + len(replacement_item.required_capitalization) :
                ]
            )
        assert new_text != text_to_check, "The new text must always be different."
        self.register_fix_token_request(
            context, token, "next_token", name_to_check, new_text
        )

    def __apply_matching_replacement(
        self,
        context: PluginScanContext,
        token: MarkdownToken,
        text_to_check: Optional[str],
        name_to_check: str,
    ) -> None:
        assert text_to_check is not None
        new_text = text_to_check
        for replacement_item in self.__replacement_items:
            if replacement_item.part_context == name_to_check:
                new_text = (
                    new_text[: replacement_item.found_index]
                    + replacement_item.required_capitalization
                    + new_text[
                        replacement_item.found_index
                        + len(replacement_item.required_capitalization) :
                    ]
                )
        if new_text != text_to_check:
            self.register_fix_token_request(
                context, token, "next_token", name_to_check, new_text
            )



"""
Module to implement a plugin that looks for heading that increment more than one
level at a time (going up).
"""

from typing import List, cast

from pymarkdown.extensions.front_matter_markdown_token import FrontMatterMarkdownToken
from pymarkdown.plugin_manager.plugin_details import (
    PluginDetailsV2,
    PluginDetailsV3,
    QueryConfigItem,
)
from pymarkdown.plugin_manager.plugin_scan_context import PluginScanContext
from pymarkdown.plugin_manager.rule_plugin import RulePlugin
from pymarkdown.tokens.markdown_token import MarkdownToken
from pymarkdown.tokens.setext_heading_markdown_token import SetextHeadingMarkdownToken


class RuleMd001(RulePlugin):
    """
    Class to implement a plugin that looks for headings that increment more than one
    level at a time (going up).
    """

    def __init__(self) -> None:
        super().__init__()
        self.__last_heading_count: int = 0
        self.__front_matter_title: str = ""

    def get_details(self) -> PluginDetailsV2:
        """
        Get the details for the plugin.
        """
        return PluginDetailsV3(
            plugin_name="heading-increment,header-increment",
            plugin_id="MD001",
            plugin_enabled_by_default=True,
            plugin_description="Heading levels should only increment by one level at a time.",
            plugin_version="0.6.0",
            plugin_url="https://pymarkdown.readthedocs.io/en/latest/plugins/rule_md001.md",
            plugin_configuration="front_matter_title",
            plugin_supports_fix=True,
        )

    def initialize_from_config(self) -> None:
        self.__front_matter_title = (
            self.plugin_configuration.get_string_property_with_default(
                "front_matter_title", "title"
            )
        )

    def query_config(self) -> List[QueryConfigItem]:
        """
        Query to find out the configuration that the rule is using.
        """
        return [QueryConfigItem("front_matter_title", self.__front_matter_title)]

    def starting_new_file(self) -> None:
        """
        Event that the a new file to be scanned is starting.
        """
        self.__last_heading_count = 0

    def next_token(self, context: PluginScanContext, token: MarkdownToken) -> None:
        """
        Event that a new token is being processed.
        """
        hash_count = None
        if token.is_atx_heading or token.is_setext_heading:
            setext_token = cast(SetextHeadingMarkdownToken, token)
            hash_count = setext_token.hash_count
        elif token.is_front_matter:
            front_matter_token = cast(FrontMatterMarkdownToken, token)
            if self.__front_matter_title in front_matter_token.matter_map:
                hash_count = 1

        if hash_count:
            if self.__last_heading_count and (hash_count > self.__last_heading_count):
                delta = hash_count - self.__last_heading_count
                if delta > 1:
                    if context.in_fix_mode:
                        hash_count = self.__last_heading_count + 1
                        self.register_fix_token_request(
                            context, token, "next_token", "hash_count", hash_count
                        )
                    else:
                        extra_data = f"Expected: h{self.__last_heading_count + 1}; Actual: h{hash_count}"
                        self.report_next_token_error(
                            context, token, extra_error_information=extra_data
                        )
            self.__last_heading_count = hash_count



"""
Module to implement a plugin that ensures that Ordered List Items have
consistent numeric prefaces.
"""

from typing import List, Optional, Tuple, cast

from pymarkdown.plugin_manager.plugin_details import (
    PluginDetailsV2,
    PluginDetailsV3,
    QueryConfigItem,
)
from pymarkdown.plugin_manager.plugin_scan_context import PluginScanContext
from pymarkdown.plugin_manager.rule_plugin import RulePlugin
from pymarkdown.tokens.list_start_markdown_token import ListStartMarkdownToken
from pymarkdown.tokens.markdown_token import MarkdownToken


class RuleMd029(RulePlugin):
    """
    Class to implement a plugin that ensures that Ordered List Items have
    consistent numeric prefaces.
    """

    __one_style = "one"
    __ordered_style = "ordered"
    __zero_style = "zero"
    __one_or_ordered_style = "one_or_ordered"

    __valid_styles = [
        __one_style,
        __ordered_style,
        __zero_style,
        __one_or_ordered_style,
    ]

    def __init__(self) -> None:
        super().__init__()
        self.__style = ""
        self.__allow_extended_start_values = False
        self.__list_stack: List[MarkdownToken] = []
        self.__ordered_list_stack: List[Tuple[Optional[str], Optional[int]]] = []

    def get_details(self) -> PluginDetailsV2:
        """
        Get the details for the plugin.
        """
        return PluginDetailsV3(
            plugin_name="ol-prefix",
            plugin_id="MD029",
            plugin_enabled_by_default=True,
            plugin_description="Ordered list item prefix",
            plugin_version="0.6.0",
            plugin_url="https://pymarkdown.readthedocs.io/en/latest/plugins/rule_md029.md",
            plugin_configuration="style,allow_extended_start_values",
            plugin_supports_fix=True,
        )

    @classmethod
    def __validate_configuration_style(cls, found_value: str) -> None:
        if found_value not in RuleMd029.__valid_styles:
            raise ValueError(f"Allowable values: {RuleMd029.__valid_styles}")

    def initialize_from_config(self) -> None:
        """
        Event to allow the plugin to load configuration information.
        """
        self.__style = self.plugin_configuration.get_string_property_with_default(
            "style",
            RuleMd029.__one_or_ordered_style,
            valid_value_fn=self.__validate_configuration_style,
        )
        self.__allow_extended_start_values = (
            self.plugin_configuration.get_boolean_property_with_default(
                "allow_extended_start_values", False
            )
        )

        # add optional to allow non- 0-1 start for ordered

    def query_config(self) -> List[QueryConfigItem]:
        """
        Query to find out the configuration that the rule is using.
        """
        return [
            QueryConfigItem("style", self.__style),
            QueryConfigItem(
                "allow_extended_start_values", self.__allow_extended_start_values
            ),
        ]

    def starting_new_file(self) -> None:
        """
        Event that the a new file to be scanned is starting.
        """
        self.__list_stack = []
        self.__ordered_list_stack = []

    def __calculate_match_info(
        self, list_style: str, initial: bool, last_known_number: Optional[int]
    ) -> Tuple[str, int]:
        if list_style == RuleMd029.__ordered_style:
            style = "1/2/3"
            if initial:
                expected_number = 1
            else:
                assert last_known_number is not None
                expected_number = last_known_number + 1
        elif list_style == RuleMd029.__one_style:
            style = "1/1/1"
            expected_number = 1
        else:
            assert list_style == RuleMd029.__zero_style
            style = "0/0/0"
            expected_number = 0
        return style, expected_number

    def __match_first_item(
        self, context: PluginScanContext, token: MarkdownToken
    ) -> Tuple[Optional[str], Optional[int]]:
        list_token = cast(ListStartMarkdownToken, token)
        last_known_number: int = int(list_token.list_start_content)

        list_style = (
            RuleMd029.__ordered_style
            if self.__style == RuleMd029.__one_or_ordered_style
            and last_known_number != 1
            else self.__style
        )
        if list_style == RuleMd029.__ordered_style:
            is_valid = self.__allow_extended_start_values or last_known_number in {0, 1}
        elif list_style == RuleMd029.__one_style:
            is_valid = last_known_number == 1
        elif list_style == RuleMd029.__zero_style:
            is_valid = last_known_number == 0
        else:
            is_valid = True
        # print(f"list_style={list_style},last_known_number={last_known_number},is_valid={is_valid}")
        if is_valid:
            return list_style, last_known_number

        return self.__report_invalid(
            context, list_token, True, list_style, last_known_number, None
        )

    def __match_non_first_items(
        self,
        context: PluginScanContext,
        token: MarkdownToken,
        list_style: Optional[str],
        last_known_number: Optional[int],
    ) -> Tuple[Optional[str], Optional[int]]:
        if list_style:
            list_token = cast(ListStartMarkdownToken, token)
            new_number: int = int(list_token.list_start_content)

            # print(f"list_style={list_style},last_known_number={last_known_number},new_number={new_number}")
            if list_style == RuleMd029.__one_or_ordered_style:
                list_style = (
                    RuleMd029.__one_style
                    if new_number == 1
                    else RuleMd029.__ordered_style
                )

            # is_valid = False
            if list_style == RuleMd029.__one_style:
                is_valid = new_number == 1
            elif list_style == RuleMd029.__zero_style:
                is_valid = new_number == 0
            else:
                assert list_style == RuleMd029.__ordered_style
                assert last_known_number is not None
                is_valid = new_number == last_known_number + 1
            if not is_valid:
                return self.__report_invalid(
                    context,
                    list_token,
                    False,
                    list_style,
                    last_known_number,
                    new_number,
                )
            last_known_number = new_number
        return list_style, last_known_number

    # pylint: disable=too-many-arguments
    def __report_invalid(
        self,
        context: PluginScanContext,
        token: ListStartMarkdownToken,
        initial_match: bool,
        list_style: str,
        last_known_number: Optional[int],
        new_number: Optional[int],
    ) -> Tuple[Optional[str], Optional[int]]:
        style, expected_number = self.__calculate_match_info(
            list_style, initial_match, last_known_number
        )
        actual_number = last_known_number if new_number is None else new_number
        extra_error_information = (
            f"Expected: {expected_number}; Actual: {actual_number}; Style: {style}"
        )
        if context.in_fix_mode:
            self.register_fix_token_request(
                context,
                token,
                "next_token",
                "list_start_content",
                str(expected_number),
            )
            if not initial_match and new_number is not None:
                expected_number_as_string = str(expected_number)
                new_number_as_string = str(new_number)
                if delta := len(expected_number_as_string) - len(new_number_as_string):
                    self.register_fix_token_request(
                        context,
                        token,
                        "next_token",
                        "indent_level",
                        token.indent_level + delta,
                    )
            return list_style, expected_number
        self.report_next_token_error(
            context, token, extra_error_information=extra_error_information
        )
        return (None, None)

    # pylint: enable=too-many-arguments

    def next_token(self, context: PluginScanContext, token: MarkdownToken) -> None:
        """
        Event that a new token is being processed.
        """
        if token.is_list_start:
            self.__list_stack.append(token)
            if token.is_ordered_list_start:
                list_style, last_known_number = self.__match_first_item(context, token)
                self.__ordered_list_stack.append((list_style, last_known_number))
        elif token.is_list_end:
            del self.__list_stack[-1]
            if token.is_ordered_list_end:
                del self.__ordered_list_stack[-1]
        elif token.is_new_list_item and self.__list_stack[-1].is_ordered_list_start:
            list_style, last_known_number = self.__ordered_list_stack[-1]
            list_style, last_known_number = self.__match_non_first_items(
                context, token, list_style, last_known_number
            )
            self.__ordered_list_stack[-1] = (list_style, last_known_number)



"""
Module to help with the parsing of backslash inline elements.
"""

import logging

from pymarkdown.container_blocks.parse_block_pass_properties import (
    ParseBlockPassProperties,
)
from pymarkdown.general.parser_helper import ParserHelper
from pymarkdown.general.parser_logger import ParserLogger
from pymarkdown.inline.inline_character_reference_helper import (
    InlineCharacterReferenceHelper,
)
from pymarkdown.inline.inline_request import InlineRequest
from pymarkdown.inline.inline_response import InlineResponse

POGGER = ParserLogger(logging.getLogger(__name__))


class InlineBackslashHelper:
    """
    Class to help with the parsing of backslash inline elements.
    """

    backslash_character = "\\"
    __backslash_punctuation = "!\"#$%&'()*+,-./:;<=>?@[]^_`{|}~\\"

    __valid_backslash_sequence_starts = f"{backslash_character}{InlineCharacterReferenceHelper.character_reference_start_character}"

    @staticmethod
    def handle_inline_backslash(
        parser_properties: ParseBlockPassProperties,
        inline_request: InlineRequest,
        add_text_signature: bool = True,
    ) -> InlineResponse:
        """
        Handle the inline case of having a backslash.
        """

        inline_response = InlineResponse()
        (
            inline_response.new_index,
            inline_response.new_string,
            inline_response.new_string_unresolved,
        ) = (inline_request.next_index + 1, "", "")
        if (
            inline_response.new_index >= len(inline_request.source_text)
            or inline_request.source_text[inline_response.new_index]
            == ParserHelper.newline_character
        ):
            inline_response.new_string, inline_response.new_string_unresolved = (
                InlineBackslashHelper.backslash_character,
                inline_response.new_string,
            )
        else:
            if (
                inline_request.source_text[inline_response.new_index]
                in InlineBackslashHelper.__backslash_punctuation
            ):
                inline_response.new_string = (
                    ParserHelper.backslash_escape_sequence if add_text_signature else ""
                ) + inline_request.source_text[inline_response.new_index]
                inline_response.new_string_unresolved = (
                    InlineBackslashHelper.backslash_character
                    + inline_response.new_string
                )
            else:
                inline_response.new_string = (
                    InlineBackslashHelper.backslash_character
                    + inline_request.source_text[inline_response.new_index]
                )
                inline_response.new_string_unresolved = inline_response.new_string
            inline_response.new_index += 1

        inline_response.delta_line_number, inline_response.delta_column_number = (
            0,
            inline_response.new_index - inline_request.next_index,
        )
        return inline_response

    @staticmethod
    def handle_backslashes(
        parser_properties: ParseBlockPassProperties, source_text: str
    ) -> str:
        """
        Handle the processing of backslashes for anything other than the text
        blocks, which have additional needs for parsing.
        """

        start_index, string_parts = 0, []
        next_index = ParserHelper.index_any_of(
            source_text,
            InlineBackslashHelper.__valid_backslash_sequence_starts,
            start_index,
        )
        while next_index != -1:
            string_parts.append(source_text[start_index:next_index])
            current_char = source_text[next_index]

            inline_request = InlineRequest(source_text, next_index)
            POGGER.debug("handle_backslashes>>$>>", current_char)
            if current_char == InlineBackslashHelper.backslash_character:
                inline_response = InlineBackslashHelper.handle_inline_backslash(
                    parser_properties, inline_request, add_text_signature=False
                )
            else:
                assert (
                    source_text[next_index]
                    == InlineCharacterReferenceHelper.character_reference_start_character
                ), "If not a backslahs, must be a character reference character."
                inline_response = (
                    InlineCharacterReferenceHelper.handle_character_reference(
                        parser_properties, inline_request
                    )
                )
            new_string, new_index = (
                inline_response.new_string,
                inline_response.new_index,
            )
            assert (
                new_string is not None and new_index is not None
            ), "String and index results must be valid."
            POGGER.debug("handle_backslashes<<$<<$", new_string, new_index)
            string_parts.append(new_string)
            start_index = new_index
            next_index = ParserHelper.index_any_of(
                source_text,
                InlineBackslashHelper.__valid_backslash_sequence_starts,
                start_index,
            )

        if start_index < len(source_text):
            string_parts.append(source_text[start_index:])
        return "".join(string_parts)



"""
Module to provide for the ability to search the text for a link.
"""

import logging
from typing import List, Optional, Tuple, cast

from pymarkdown.container_blocks.parse_block_pass_properties import (
    ParseBlockPassProperties,
)
from pymarkdown.general.constants import Constants
from pymarkdown.general.parser_helper import ParserHelper
from pymarkdown.general.parser_logger import ParserLogger
from pymarkdown.inline.emphasis_helper import EmphasisHelper
from pymarkdown.links.link_create_helper import (
    LinkCreateHelper,
    ProcessSimpleInlineProtocol,
)
from pymarkdown.links.link_helper_properties import LinkHelperProperties
from pymarkdown.links.link_parse_helper import LinkParseHelper
from pymarkdown.tokens.email_autolink_markdown_token import EmailAutolinkMarkdownToken
from pymarkdown.tokens.hard_break_markdown_token import HardBreakMarkdownToken
from pymarkdown.tokens.image_start_markdown_token import ImageStartMarkdownToken
from pymarkdown.tokens.inline_code_span_markdown_token import (
    InlineCodeSpanMarkdownToken,
)
from pymarkdown.tokens.link_start_markdown_token import LinkStartMarkdownToken
from pymarkdown.tokens.markdown_token import MarkdownToken
from pymarkdown.tokens.raw_html_markdown_token import RawHtmlMarkdownToken
from pymarkdown.tokens.special_text_markdown_token import SpecialTextMarkdownToken
from pymarkdown.tokens.text_markdown_token import TextMarkdownToken

POGGER = ParserLogger(logging.getLogger(__name__))


# pylint: disable=too-few-public-methods


class LinkSearchHelper:
    """
    Class to provide for the ability to search the text for a link.
    """

    __link_start_sequence = "["
    image_start_sequence = "!["
    __valid_link_starts = [__link_start_sequence, image_start_sequence]

    # pylint: disable=too-many-arguments, too-many-locals
    @staticmethod
    def look_for_link_or_image(
        parser_properties: ParseBlockPassProperties,
        inline_blocks: List[MarkdownToken],
        source_text: str,
        next_index: int,
        remaining_line: str,
        tabified_remaining_line: Optional[str],
        current_string_unresolved: str,
        process_inlines_fn: ProcessSimpleInlineProtocol,
        tabified_text: Optional[str],
    ) -> Tuple[int, bool, Optional[MarkdownToken], bool]:
        """
        Given that a link close character has been found, process it to see if
        there is actually enough other text to properly construct the link.
        """

        POGGER.debug(
            ">>look_for_link_or_image>>$<<",
            inline_blocks,
        )
        POGGER.debug(">>source_text>>$<<", source_text)
        POGGER.debug(">>tabified_text>>$<<", tabified_text)
        POGGER.debug(">>next_index>>$<<", next_index)
        POGGER.debug(">>remaining_line>>$<<", remaining_line)
        POGGER.debug(">>tabified_remaining_line>>$<<", tabified_remaining_line)
        POGGER.debug(
            ">>current_string_unresolved>>$<<",
            current_string_unresolved,
        )
        is_valid, consume_rest_of_line, next_index, updated_index = (
            False,
            False,
            next_index + 1,
            -1,
        )
        token_to_append: Optional[MarkdownToken] = None

        POGGER.debug("LOOKING FOR START")
        LinkSearchHelper.__debug_log_specials_in_tokens(inline_blocks)

        valid_special_start_text, search_index = None, len(inline_blocks) - 1
        while search_index >= 0:
            if inline_blocks[search_index].is_special_text:
                (
                    is_done,
                    is_valid,
                    updated_index,
                    token_to_append,
                    consume_rest_of_line,
                    valid_special_start_text,
                ) = LinkSearchHelper.__find_link(
                    parser_properties,
                    inline_blocks,
                    search_index,
                    source_text,
                    next_index,
                    remaining_line,
                    tabified_remaining_line,
                    current_string_unresolved,
                    process_inlines_fn,
                    updated_index,
                    tabified_text,
                )
                if is_done:
                    break
                POGGER.debug("  not link")
            search_index -= 1

        return LinkSearchHelper.__look_for_link_or_image_end(
            inline_blocks,
            is_valid,
            consume_rest_of_line,
            valid_special_start_text,
            search_index,
            next_index,
            updated_index,
            token_to_append,
        )

    # pylint: enable=too-many-arguments, too-many-locals

    # pylint: disable=too-many-arguments
    @staticmethod
    def __look_for_link_or_image_end(
        inline_blocks: List[MarkdownToken],
        is_valid: bool,
        consume_rest_of_line: bool,
        valid_special_start_text: Optional[str],
        search_index: int,
        next_index: int,
        updated_index: int,
        token_to_append: Optional[MarkdownToken],
    ) -> Tuple[int, bool, Optional[MarkdownToken], bool]:
        POGGER.debug(
            ">>look_for_link_or_image>>$<<is_valid<<$<<$<<",
            inline_blocks,
            is_valid,
            consume_rest_of_line,
        )
        if is_valid:
            assert (
                valid_special_start_text is not None
            ), "if is_valid is True, valid_special_start_text must be defined"
            LinkSearchHelper.__deactivate_used_tokens(
                inline_blocks, search_index, valid_special_start_text
            )
            return updated_index, True, token_to_append, consume_rest_of_line
        return next_index, False, token_to_append, consume_rest_of_line

    # pylint: enable=too-many-arguments

    # pylint: disable=too-many-arguments, too-many-locals
    @staticmethod
    def __find_link(
        parser_properties: ParseBlockPassProperties,
        inline_blocks: List[MarkdownToken],
        search_index: int,
        source_text: str,
        new_index: int,
        remaining_line: str,
        tabified_remaining_line: Optional[str],
        current_string_unresolved: str,
        process_inlines_fn: ProcessSimpleInlineProtocol,
        updated_index: int,
        tabified_text: Optional[str],
    ) -> Tuple[bool, bool, int, Optional[MarkdownToken], bool, Optional[str]]:
        special_text_token = cast(SpecialTextMarkdownToken, inline_blocks[search_index])
        # POGGER.debug(
        #     "search_index>>$>>$",
        #     search_index,
        #     special_text_token.show_process_emphasis(),
        # )

        (
            is_done,
            consume_rest_of_line,
            is_valid,
        ) = (False, False, False)
        token_to_append: Optional[MarkdownToken] = None
        if special_text_token.token_text in LinkSearchHelper.__valid_link_starts:
            valid_special_start_text: Optional[str] = special_text_token.token_text
            if special_text_token.is_active:
                # POGGER.debug(">>>>>>$", inline_blocks)
                assert (
                    valid_special_start_text is not None
                ), "if active, the token_text must be defined"
                (
                    updated_index,
                    token_to_append,
                    consume_rest_of_line,
                ) = LinkSearchHelper.__handle_link_types(
                    parser_properties,
                    inline_blocks,
                    search_index,
                    source_text,
                    new_index,
                    valid_special_start_text,
                    remaining_line,
                    tabified_remaining_line,
                    current_string_unresolved,
                    process_inlines_fn,
                    tabified_text,
                )
                if updated_index != -1:
                    is_valid, is_done = True, True

            if not is_done:
                # POGGER.debug("  not active:$", search_index)
                LinkSearchHelper.__revert_token_to_normal_text_token(
                    inline_blocks, search_index
                )
                is_done = True
        else:
            valid_special_start_text = None
        return (
            is_done,
            is_valid,
            updated_index,
            token_to_append,
            consume_rest_of_line,
            valid_special_start_text,
        )

    # pylint: enable=too-many-arguments, too-many-locals
    @staticmethod
    def __debug_log_specials_in_tokens(inline_blocks: List[MarkdownToken]) -> None:
        display_parts = []
        for deactivate_token in inline_blocks:
            if deactivate_token.is_special_text:
                special_token = cast(SpecialTextMarkdownToken, deactivate_token)
                display_parts.extend(
                    [
                        f",>>Spec:{special_token.is_active}:{deactivate_token}<<",
                    ]
                )
            else:
                display_parts.extend([",", str(deactivate_token)])

        POGGER.debug("$", "".join(display_parts)[1:])

    @staticmethod
    def __revert_token_to_normal_text_token(
        inline_blocks: List[MarkdownToken], search_index: int
    ) -> None:
        """
        Revert this token from a special text token back to a normal text token.
        """

        POGGER.debug("REVERTING")
        LinkSearchHelper.__debug_log_specials_in_tokens(inline_blocks)

        text_token_to_replace = cast(
            SpecialTextMarkdownToken, inline_blocks[search_index]
        )
        inline_blocks.insert(search_index, text_token_to_replace.create_copy())
        del inline_blocks[search_index + 1]

        POGGER.debug("REVERTED")
        LinkSearchHelper.__debug_log_specials_in_tokens(inline_blocks)

    @staticmethod
    def __deactivate_used_tokens(
        inline_blocks: List[MarkdownToken],
        search_index: int,
        valid_special_start_text: str,
    ) -> None:
        # if link set all [ before to inactive
        POGGER.debug("")
        POGGER.debug("SET TO INACTIVE-->$", valid_special_start_text)
        POGGER.debug("ind-->$", search_index)

        assert (
            inline_blocks[search_index].is_inline_link
            or inline_blocks[search_index].is_inline_image
        ), "token to deactivate must be a link or image"

        POGGER.debug(
            "\nresolve_inline_emphasis>>$",
            inline_blocks,
        )
        EmphasisHelper.resolve_inline_emphasis(
            inline_blocks, inline_blocks[search_index]
        )
        POGGER.debug(
            "resolve_inline_emphasis>>$\n",
            inline_blocks,
        )

        if valid_special_start_text == LinkSearchHelper.__link_start_sequence:
            POGGER.debug("DEACTIVATING")
            LinkSearchHelper.__debug_log_specials_in_tokens(inline_blocks)
            for deactivate_token in inline_blocks:
                if deactivate_token.is_special_text:
                    special_token = cast(SpecialTextMarkdownToken, deactivate_token)
                    POGGER.debug("inline_blocks>>>>>>>>>>>>>>>>>>$", special_token)
                    if (
                        special_token.token_text
                        == LinkSearchHelper.__link_start_sequence
                    ):
                        special_token.deactivate()
            POGGER.debug("DEACTIVATED")
            LinkSearchHelper.__debug_log_specials_in_tokens(inline_blocks)

    # pylint: disable=too-many-arguments
    @staticmethod
    def __handle_link_types(
        parser_properties: ParseBlockPassProperties,
        inline_blocks: List[MarkdownToken],
        ind: int,
        source_text: str,
        new_index: int,
        start_text: str,
        remaining_line: str,
        tabified_remaining_line: Optional[str],
        current_string_unresolved: str,
        process_inlines_fn: ProcessSimpleInlineProtocol,
        tabified_text: Optional[str],
    ) -> Tuple[int, Optional[MarkdownToken], bool]:
        """
        After finding a link specifier, figure out what type of link it is.
        """

        (
            update_index,
            text_from_blocks,
            text_from_blocks_raw,
            lhp,
        ) = LinkSearchHelper.__excavate_link(
            parser_properties,
            inline_blocks,
            ind,
            source_text,
            new_index,
            current_string_unresolved,
            remaining_line,
            tabified_remaining_line,
            tabified_text,
        )

        POGGER.debug("<<<<<<<new_index<<<<<<<$<<", new_index)
        POGGER.debug("<<<<<<<update_index<<<<<<<$<<", update_index)
        POGGER.debug("<<<<<<<text_from_blocks_raw<<<<<<<$<<", text_from_blocks_raw)
        if update_index != -1:
            assert (
                lhp.label_type is not None
                and lhp.before_link_whitespace is not None
                and lhp.after_title_whitespace is not None
                and lhp.before_title_whitespace is not None
                and lhp.inline_title is not None
                and lhp.ex_label is not None
                and lhp.inline_link is not None
                and lhp.pre_inline_link is not None
                and lhp.pre_inline_title is not None
                and lhp.bounding_character is not None
            ), "if successful, this field must be defined"
            return LinkCreateHelper.create_link_token(
                parser_properties,
                start_text,
                text_from_blocks,
                text_from_blocks_raw,
                inline_blocks,
                ind,
                remaining_line,
                current_string_unresolved,
                process_inlines_fn,
                lhp,
                update_index,
            )
        return update_index, None, False

    # pylint: enable=too-many-arguments

    # pylint: disable=too-many-arguments
    @staticmethod
    def __excavate_link(
        parser_properties: ParseBlockPassProperties,
        inline_blocks: List[MarkdownToken],
        ind: int,
        source_text: str,
        new_index: int,
        current_string_unresolved: str,
        remaining_line: str,
        tabified_remaining_line: Optional[str],
        tabified_text: Optional[str],
    ) -> Tuple[
        int,
        str,
        str,
        LinkHelperProperties,
    ]:
        POGGER.debug(
            "handle_link_types>>$<<$<<",
            inline_blocks,
            ind,
        )
        POGGER.debug(
            "handle_link_types>source_text>>$<<",
            source_text,
        )
        POGGER.debug("handle_link_types>>$<<", source_text[new_index:])
        POGGER.debug(
            "handle_link_types>>current_string_unresolved>:$:<remaining_line>:$:<tabified_remaining_line>:$:<",
            current_string_unresolved,
            remaining_line,
            tabified_remaining_line,
        )
        text_to_scan = remaining_line
        if tabified_remaining_line is not None:
            text_to_scan = tabified_remaining_line
        (
            text_from_blocks,
            text_from_blocks_raw,
        ) = LinkSearchHelper.__collect_text_from_blocks(
            inline_blocks, ind, f"{current_string_unresolved}{text_to_scan}"
        )
        POGGER.debug(
            "handle_link_types>>text_from_blocks_raw>>$<<",
            text_from_blocks_raw,
        )
        POGGER.debug(
            "handle_link_types>>text_from_blocks>>$<<",
            text_from_blocks,
        )

        POGGER.debug("__look_for_link_formats>>$>>", new_index)
        (
            update_index,
            tried_full_reference_form,
            lhp,
        ) = LinkParseHelper.look_for_link_formats(
            parser_properties, source_text, new_index, text_from_blocks, tabified_text
        )

        # u != -1 - inline valid
        # tried_full_reference_form - collapsed or full valid
        if update_index == -1 and not tried_full_reference_form:
            (
                lhp.ex_label,
                update_index,
                lhp.inline_link,
                lhp.inline_title,
                lhp.label_type,
                lhp.pre_inline_link,
            ) = LinkSearchHelper.__look_for_shortcut_link(
                inline_blocks, text_from_blocks, new_index
            )
        return (
            update_index,
            text_from_blocks,
            text_from_blocks_raw,
            lhp,
        )

    # pylint: enable=too-many-arguments
    @staticmethod
    def __look_for_shortcut_link(
        inline_blocks: List[MarkdownToken], text_from_blocks: str, new_index: int
    ) -> Tuple[str, int, str, str, str, str]:
        POGGER.debug("shortcut?")
        POGGER.debug(
            ">>$<<",
            inline_blocks,
        )
        POGGER.debug(
            ">>$<<",
            text_from_blocks,
        )

        update_index, inline_link, inline_title = LinkParseHelper.look_up_link(
            text_from_blocks, new_index, Constants.link_type__shortcut
        )
        return (
            "",
            update_index,
            inline_link,
            inline_title,
            Constants.link_type__shortcut,
            "",
        )

    @staticmethod
    def __collect_text_from_blocks(
        inline_blocks: List[MarkdownToken], ind: int, suffix_text: str
    ) -> Tuple[str, str]:
        """
        Aggregate the text component of text blocks.
        """

        # POGGER.debug(
        #     ">>collect_text_from_blocks>>$",
        #     inline_blocks,
        # )
        POGGER.debug(
            ">>collect_text_from_blocks>>suffix_text>>$",
            suffix_text,
        )

        (
            collect_index,
            is_inside_of_link,
            inline_blocks_size,
        ) = (ind + 1, False, len(inline_blocks))
        text_parts: List[str] = []
        text_raw_parts: List[str] = []
        POGGER.debug(
            ">>collect_index:$: < inline_blocks_size:$:",
            collect_index,
            inline_blocks_size,
        )
        while collect_index < inline_blocks_size:
            (
                is_inside_of_link,
                collect_index,
            ) = LinkSearchHelper.__collect_text_from_next_block(
                inline_blocks,
                inline_blocks_size,
                collect_index,
                text_raw_parts,
                text_parts,
                is_inside_of_link,
            )

        text_parts.append(suffix_text)
        text_raw_parts.append(suffix_text)

        text_from_blocks = "".join(text_parts)
        POGGER.debug(">>text_from_blocks>:$:<", text_from_blocks)
        text_from_blocks = ParserHelper.resolve_backspaces_from_text(text_from_blocks)
        POGGER.debug(">>text_from_blocks>:$:<", text_from_blocks)
        POGGER.debug(">>text_raw_parts>:$:<", text_raw_parts)
        return text_from_blocks, "".join(text_raw_parts)

    # pylint: disable=too-many-arguments
    @staticmethod
    def __collect_text_from_next_block(
        inline_blocks: List[MarkdownToken],
        inline_blocks_size: int,
        collect_index: int,
        text_raw_parts: List[str],
        text_parts: List[str],
        is_inside_of_link: bool,
    ) -> Tuple[bool, int]:
        POGGER.debug(">>collect_text>>$<<", inline_blocks[collect_index])
        if inline_blocks[collect_index].is_inline_link_end:
            is_inside_of_link = LinkSearchHelper.__collect_text_from_inline_link_end()
        elif inline_blocks[collect_index].is_inline_link:
            is_inside_of_link = LinkSearchHelper.__collect_text_from_inline_link_start(
                inline_blocks, collect_index, text_raw_parts
            )
        elif inline_blocks[collect_index].is_inline_image:
            LinkSearchHelper.__collect_text_from_inline_image(
                inline_blocks, collect_index, text_raw_parts, text_parts
            )
        elif inline_blocks[collect_index].is_inline_code_span:
            LinkSearchHelper.__collect_text_from_inline_code_span(
                inline_blocks,
                collect_index,
                text_raw_parts,
                text_parts,
                is_inside_of_link,
            )
        elif inline_blocks[collect_index].is_inline_raw_html:
            LinkSearchHelper.__collect_text_from_inline_raw_html(
                inline_blocks,
                collect_index,
                text_raw_parts,
                text_parts,
                is_inside_of_link,
            )
        elif inline_blocks[collect_index].is_inline_autolink:
            LinkSearchHelper.__collect_text_from_inline_autolink(
                inline_blocks,
                collect_index,
                text_raw_parts,
                text_parts,
                is_inside_of_link,
            )
        elif inline_blocks[collect_index].is_inline_hard_break:
            LinkSearchHelper.__collect_text_from_inline_hard_break(
                inline_blocks, collect_index, text_raw_parts, text_parts
            )
        elif not is_inside_of_link:
            LinkSearchHelper.__collect_text_from_text(
                inline_blocks, collect_index, text_raw_parts, text_parts
            )
        POGGER.debug(">>collect_text>>$<<$<<", text_parts, inline_blocks[collect_index])
        POGGER.debug(">>collected_text_raw>>$<<", text_raw_parts)
        collect_index += 1
        POGGER.debug(
            ">>collect_index:$: < inline_blocks_size:$:",
            collect_index,
            inline_blocks_size,
        )
        return is_inside_of_link, collect_index

    # pylint: enable=too-many-arguments

    @staticmethod
    def __collect_text_from_inline_link_end() -> bool:
        return False

    @staticmethod
    def __collect_text_from_inline_link_start(
        inline_blocks: List[MarkdownToken],
        collect_index: int,
        text_raw_parts: List[str],
    ) -> bool:
        link_token = cast(LinkStartMarkdownToken, inline_blocks[collect_index])
        raw_text = LinkStartMarkdownToken.rehydrate_inline_link_text_from_token(
            link_token
        )
        text_raw_parts.append(raw_text)
        return True

    @staticmethod
    def __collect_text_from_inline_image(
        inline_blocks: List[MarkdownToken],
        collect_index: int,
        text_raw_parts: List[str],
        text_parts: List[str],
    ) -> None:
        image_token = cast(ImageStartMarkdownToken, inline_blocks[collect_index])
        text_raw_parts.append(
            ImageStartMarkdownToken.rehydrate_inline_image_text_from_token(image_token)
        )
        text_parts.append(image_token.image_alt_text)

    @staticmethod
    def __collect_text_from_inline_code_span(
        inline_blocks: List[MarkdownToken],
        collect_index: int,
        text_raw_parts: List[str],
        text_parts: List[str],
        is_inside_of_link: bool,
    ) -> None:
        if not is_inside_of_link:
            codespan_token = cast(
                InlineCodeSpanMarkdownToken, inline_blocks[collect_index]
            )
            POGGER.debug("CODESPAN>>$<<", codespan_token)
            sub_parts = [
                codespan_token.extracted_start_backticks,
                ParserHelper.remove_all_from_text(codespan_token.leading_whitespace),
                ParserHelper.remove_all_from_text(codespan_token.span_text),
                ParserHelper.remove_all_from_text(codespan_token.trailing_whitespace),
                codespan_token.extracted_start_backticks,
            ]
            text_parts.extend(sub_parts)
            text_raw_parts.extend(sub_parts)

    @staticmethod
    def __collect_text_from_inline_raw_html(
        inline_blocks: List[MarkdownToken],
        collect_index: int,
        text_raw_parts: List[str],
        text_parts: List[str],
        is_inside_of_link: bool,
    ) -> None:
        if not is_inside_of_link:
            html_token = cast(RawHtmlMarkdownToken, inline_blocks[collect_index])
            sub_parts = ["<", html_token.raw_tag, ">"]
            text_parts.extend(sub_parts)
            text_raw_parts.extend(sub_parts)

    @staticmethod
    def __collect_text_from_inline_autolink(
        inline_blocks: List[MarkdownToken],
        collect_index: int,
        text_raw_parts: List[str],
        text_parts: List[str],
        is_inside_of_link: bool,
    ) -> None:
        if not is_inside_of_link:
            autolink_token = cast(
                EmailAutolinkMarkdownToken, inline_blocks[collect_index]
            )
            sub_parts = ["<", autolink_token.autolink_text, ">"]
            text_parts.extend(sub_parts)
            text_raw_parts.extend(sub_parts)

    @staticmethod
    def __collect_text_from_inline_hard_break(
        inline_blocks: List[MarkdownToken],
        collect_index: int,
        text_raw_parts: List[str],
        text_parts: List[str],
    ) -> None:
        break_token = cast(HardBreakMarkdownToken, inline_blocks[collect_index])
        POGGER.debug("is_inline_hard_break>>collected_text_raw>>$<<", text_raw_parts)
        converted_text = break_token.line_end
        text_parts.append(converted_text)
        if converted_text == "\\":
            text_parts.append(converted_text)
        text_parts.append(ParserHelper.newline_character)
        text_raw_parts.extend((converted_text, ParserHelper.newline_character))
        POGGER.debug("is_inline_hard_break>>collected_text_raw>>$<<", text_raw_parts)

    @staticmethod
    def __collect_text_from_text(
        inline_blocks: List[MarkdownToken],
        collect_index: int,
        text_raw_parts: List[str],
        text_parts: List[str],
    ) -> None:
        text_token = cast(TextMarkdownToken, inline_blocks[collect_index])
        text_parts.append(text_token.token_text)
        text_raw_parts.append(text_token.token_text)


# pylint: enable=too-few-public-methods



"""
Module to contain information about a failure reported by one of the rule plugins.
"""

from dataclasses import dataclass
from typing import Optional

# pylint: disable=too-many-instance-attributes


@dataclass(frozen=True)
class PluginScanFailure:
    """
    Class to contain information about a failure reported by one of the rule plugins.
    """

    scan_file: str
    line_number: int
    column_number: int
    rule_id: str
    rule_name: str
    rule_description: str
    extra_error_information: Optional[str]
    is_error_token_prefaced_by_blank_line: bool

    def __lt__(self, other: "PluginScanFailure") -> bool:
        # This is required to allow for sorting of the failures before display.

        # if self.scan_file != other.scan_file:
        #     return self.scan_file < other.scan_file
        if self.line_number != other.line_number:
            return self.line_number < other.line_number
        if self.column_number != other.column_number:
            return self.column_number < other.column_number
        return self.rule_id < other.rule_id


# pylint: enable=too-many-instance-attributes



"""
Module to implement a plugin that looks for hard tabs in the files.
"""

from typing import List, Optional, cast

from pymarkdown.general.tab_helper import TabHelper
from pymarkdown.plugin_manager.plugin_details import (
    PluginDetails,
    PluginDetailsV3,
    QueryConfigItem,
)
from pymarkdown.plugin_manager.plugin_scan_context import PluginScanContext
from pymarkdown.plugin_manager.rule_plugin import RulePlugin
from pymarkdown.tokens.fenced_code_block_markdown_token import (
    FencedCodeBlockMarkdownToken,
)
from pymarkdown.tokens.markdown_token import EndMarkdownToken, MarkdownToken


class RuleMd010(RulePlugin):
    """
    Class to implement a plugin that looks for hard tabs in the files.
    """

    def __init__(self) -> None:
        super().__init__()
        self.__leaf_tokens: List[MarkdownToken] = []
        self.__line_index = -1
        self.__leaf_token_index = -1
        self.__check_in_code_blocks: bool = False
        self.__last_fenced_code_end_token: Optional[EndMarkdownToken] = None
        self.__last_line: Optional[str] = None

    def get_details(self) -> PluginDetails:
        """
        Get the details for the plugin.
        """
        return PluginDetailsV3(
            plugin_name="no-hard-tabs",
            plugin_id="MD010",
            plugin_enabled_by_default=True,
            plugin_description="Hard tabs",
            plugin_version="0.6.1",
            plugin_url="https://pymarkdown.readthedocs.io/en/latest/plugins/rule_md010.md",
            plugin_configuration="code_blocks",
            plugin_supports_fix=True,
            plugin_fix_level=0,
        )

    def initialize_from_config(self) -> None:
        """
        Event to allow the plugin to load configuration information.
        """
        self.__check_in_code_blocks = (
            self.plugin_configuration.get_boolean_property_with_default(
                "code_blocks",
                True,
            )
        )

    def query_config(self) -> List[QueryConfigItem]:
        """
        Query to find out the configuration that the rule is using.
        """
        return [QueryConfigItem("code_blocks", self.__check_in_code_blocks)]

    def starting_new_file(self) -> None:
        """
        Event that the a new file to be scanned is starting.
        """
        # self.__container_tokens = []
        # self.__end_container_tokens_to_match :List[EndMarkdownToken] = []
        # self.__container_end_line_numbers = {}
        self.__leaf_tokens = []
        self.__line_index = 1
        self.__leaf_token_index = 0
        self.__last_fenced_code_end_token = None
        self.__last_line = None

    def next_token(self, context: PluginScanContext, token: MarkdownToken) -> None:
        """
        Event that a new token is being processed.
        """
        _ = context
        # if token.is_block_quote_start or token.is_list_start:
        #     container_token = cast(ContainerMarkdownToken, token)
        #     self.__container_tokens.append(container_token)
        # elif token.is_block_quote_end or token.is_list_end:
        #     end_token = cast(EndMarkdownToken, token)
        #     self.__end_container_tokens_to_match.append(end_token)
        # elif self.__end_container_tokens_to_match:
        #     while self.__end_container_tokens_to_match:
        #         next_end_token = self.__end_container_tokens_to_match[0]
        #         del self.__end_container_tokens_to_match[0]
        #         self.__container_end_line_numbers[next_end_token.start_markdown_token] = token.line_number

        if token.is_blank_line or token.is_leaf or token.is_end_of_stream:
            self.__leaf_tokens.append(token)
        elif token.is_fenced_code_block_end:
            self.__last_fenced_code_end_token = cast(EndMarkdownToken, token)

    def __is_line_inside_of_fenced_code_block(self) -> bool:
        is_inside_of_fenced_code_block = False
        leaf_token = self.__leaf_tokens[self.__leaf_token_index]
        if leaf_token.is_fenced_code_block:
            fenced_leaf_token = cast(FencedCodeBlockMarkdownToken, leaf_token)
            last_line_number = self.__leaf_tokens[
                self.__leaf_token_index + 1
            ].line_number
            end_code_block_line_number = (
                last_line_number
                if self.__last_fenced_code_end_token
                and self.__last_fenced_code_end_token.was_forced
                else last_line_number - 1
            )
            is_inside_of_fenced_code_block = (
                self.__line_index > fenced_leaf_token.line_number
                and self.__line_index < end_code_block_line_number
            )
        return is_inside_of_fenced_code_block

    def next_line(self, context: PluginScanContext, line: str) -> None:
        """
        Event that a new line is being processed.
        """
        if (
            self.__leaf_token_index + 1 < len(self.__leaf_tokens)
            and self.__line_index
            == self.__leaf_tokens[self.__leaf_token_index + 1].line_number
        ):
            self.__leaf_token_index += 1

        if "\t" in line:
            is_inside_of_fenced_code_block = (
                self.__is_line_inside_of_fenced_code_block()
            )
            can_trigger = (
                not self.__check_in_code_blocks and not is_inside_of_fenced_code_block
            )
            do_process = self.__check_in_code_blocks or can_trigger
            # if not do_process and not xx:
            #     ff = 0
            #     fj = []
            #     while ff < len(self.__container_tokens):
            #         fg = self.__container_tokens[ff]
            #         if self.__line_index < fg.line_number:
            #             break
            #         fh = self.__container_end_line_numbers[fg]
            #         if self.__line_index >fg.line_number and self.__line_index < fh:
            #             fj.append(fg)
            #         ff += 1
            #     x = 1
            #     assert False
        else:
            do_process = False
        if do_process:
            if context.in_fix_mode:
                context.set_current_fix_line(TabHelper.detabify_string(line))
            else:
                next_index = line.find("\t", 0)
                while next_index != -1:
                    line_before_tab = line[:next_index]
                    untabified_line_before_tab = TabHelper.detabify_string(
                        line_before_tab
                    )
                    column_number_of_tab = len(untabified_line_before_tab) + 1
                    override_is_error_token_prefaced_by_blank_line = (
                        self.__last_line is not None and not self.__last_line
                    )
                    self.report_next_line_error(
                        context,
                        column_number_of_tab,
                        extra_error_information=f"Column: {column_number_of_tab}",
                        override_is_error_token_prefaced_by_blank_line=override_is_error_token_prefaced_by_blank_line,
                    )
                    next_index = line.find("\t", next_index + 1)

        self.__line_index += 1
        self.__last_line = line.strip(" \t")



"""
Module to provide processing for the list blocks.
"""

import logging
from typing import List, Optional, Tuple, cast

from pymarkdown.block_quotes.block_quote_data import BlockQuoteData
from pymarkdown.container_blocks.container_grab_bag import ContainerGrabBag
from pymarkdown.general.parser_helper import ParserHelper
from pymarkdown.general.parser_logger import ParserLogger
from pymarkdown.general.parser_state import ParserState
from pymarkdown.general.position_marker import PositionMarker
from pymarkdown.general.requeue_line_info import RequeueLineInfo
from pymarkdown.general.tab_helper import TabHelper
from pymarkdown.html.html_helper import HtmlHelper
from pymarkdown.leaf_blocks.atx_leaf_block_processor import AtxLeafBlockProcessor
from pymarkdown.leaf_blocks.fenced_leaf_block_processor import FencedLeafBlockProcessor
from pymarkdown.leaf_blocks.leaf_block_processor import LeafBlockProcessor
from pymarkdown.leaf_blocks.leaf_block_processor_paragraph import (
    LeafBlockProcessorParagraph,
)
from pymarkdown.leaf_blocks.thematic_leaf_block_processor import (
    ThematicLeafBlockProcessor,
)
from pymarkdown.list_blocks.list_block_create_new_handler import (
    ListBlockCreateNewHandler,
)
from pymarkdown.list_blocks.list_block_pre_list_helper import ListBlockPreListHelper
from pymarkdown.list_blocks.list_block_starts_helper import ListBlockStartsHelper
from pymarkdown.tokens.list_start_markdown_token import ListStartMarkdownToken
from pymarkdown.tokens.markdown_token import MarkdownToken
from pymarkdown.tokens.stack_token import ListStackToken, StackToken

POGGER = ParserLogger(logging.getLogger(__name__))

# pylint: disable=too-many-lines


class ListBlockProcessor:
    """
    Class to provide processing for the list blocks.
    """

    @staticmethod
    def __handle_list_with_leading_indent_loop(
        acceptable_indent_stack_index: int,
        parser_state: ParserState,
        stack_index: int,
        extracted_whitespace: str,
    ) -> int:
        box_start = 1 if stack_index == 1 else ((stack_index + 1) * 4) + stack_index
        box_end = (stack_index * 4) + (stack_index - 1)

        is_in_this_box = box_start <= len(extracted_whitespace) + 1 <= box_end
        can_promote_to_next_box = (stack_index + 1) < len(
            parser_state.token_stack
        ) and parser_state.token_stack[stack_index + 1].is_list

        if is_in_this_box or can_promote_to_next_box:
            acceptable_indent_stack_index = stack_index
        return acceptable_indent_stack_index

    # pylint: disable=too-many-arguments
    @staticmethod
    def __handle_list_with_leading_indent(
        parser_state: ParserState,
        container_depth: int,
        removed_chars_at_start: int,
        extracted_whitespace: str,
        indent_already_processed: bool,
        adj_ws: str,
    ) -> Tuple[int, Optional[str], str, str, bool]:
        indent_already_used, forced_container_whitespace = 0, None
        is_in_root_list = (
            not container_depth
            and parser_state.token_stack
            and len(parser_state.token_stack) >= 2
            and parser_state.token_stack[1].is_list
        )

        if (
            not removed_chars_at_start
            and is_in_root_list
            and adj_ws == extracted_whitespace
            and len(extracted_whitespace) >= 4
            and not indent_already_processed
        ):
            stack_index = 1
            acceptable_indent_stack_index = 0
            while (
                stack_index < len(parser_state.token_stack)
                and parser_state.token_stack[stack_index].is_list
            ):
                acceptable_indent_stack_index = (
                    ListBlockProcessor.__handle_list_with_leading_indent_loop(
                        acceptable_indent_stack_index,
                        parser_state,
                        stack_index,
                        extracted_whitespace,
                    )
                )
                stack_index += 1

            if acceptable_indent_stack_index:
                list_stack_token = parser_state.token_stack[
                    acceptable_indent_stack_index
                ]
                assert list_stack_token.is_list, "Token must be a list token."
                list_token = cast(
                    ListStartMarkdownToken, list_stack_token.matching_markdown_token
                )
                indent_already_used = list_token.indent_level
                forced_container_whitespace = extracted_whitespace[:indent_already_used]
                extracted_whitespace = extracted_whitespace[indent_already_used:]
                adj_ws = adj_ws[indent_already_used:]
                indent_already_processed = True
        return (
            indent_already_used,
            forced_container_whitespace,
            extracted_whitespace,
            adj_ws,
            indent_already_processed,
        )

    # pylint: enable=too-many-arguments

    # pylint: disable=too-many-arguments
    @staticmethod
    def __handle_list_block_init(
        parser_state: ParserState,
        position_marker: PositionMarker,
        extracted_whitespace: str,
        adj_ws: str,
        is_ulist: bool,
        container_depth: int,
        removed_chars_at_start: int,
        indent_already_processed: bool,
    ) -> Tuple[
        bool,
        int,
        Optional[int],
        Optional[int],
        int,
        Optional[str],
        str,
        str,
        bool,
    ]:
        POGGER.debug(
            "hlb>>parser_state.nested_list_start>$", parser_state.nested_list_start
        )
        POGGER.debug("hlb>>extracted_whitespace>$<", extracted_whitespace)
        POGGER.debug("hlb>>adj_ws>$<", adj_ws)
        POGGER.debug("hlb>>removed_chars_at_start>$<", removed_chars_at_start)
        POGGER.debug(
            "text_to_parse[index=$:]>:$:<",
            position_marker.index_number,
            position_marker.text_to_parse[position_marker.index_number :],
        )

        (
            indent_already_used,
            forced_container_whitespace,
            extracted_whitespace,
            adj_ws,
            indent_already_processed,
        ) = ListBlockProcessor.__handle_list_with_leading_indent(
            parser_state,
            container_depth,
            removed_chars_at_start,
            extracted_whitespace,
            indent_already_processed,
            adj_ws,
        )

        if is_ulist:
            (
                started_ulist,
                end_of_ulist_start_index,
                index,
                number_of_digits,
            ) = ListBlockStartsHelper.is_ulist_start(
                parser_state,
                position_marker.text_to_parse,
                position_marker.index_number,
                extracted_whitespace,
                False,
                adj_ws,
            )
        else:
            (
                started_ulist,
                end_of_ulist_start_index,
                index,
                number_of_digits,
            ) = ListBlockStartsHelper.is_olist_start(
                parser_state,
                position_marker.text_to_parse,
                position_marker.index_number,
                extracted_whitespace,
                False,
                adj_ws,
            )
        return (
            started_ulist,
            end_of_ulist_start_index,
            index,
            number_of_digits,
            indent_already_used,
            forced_container_whitespace,
            extracted_whitespace,
            adj_ws,
            indent_already_processed,
        )

    # pylint: enable=too-many-arguments

    # pylint: disable=too-many-arguments, too-many-locals
    @staticmethod
    def __create_new_list_if_indented(
        parser_state: ParserState,
        position_marker: PositionMarker,
        indent_level: int,
        ws_before_marker: int,
        ws_after_marker: int,
        extracted_whitespace: str,
        forced_container_whitespace: Optional[str],
        adj_ws: str,
        adjusted_text_to_parse: Optional[str],
        index: int,
        container_level_tokens: List[MarkdownToken],
        remaining_whitespace: int,
        after_marker_ws_index: int,
        current_container_blocks: List[StackToken],
        container_depth: int,
        original_line: str,
        is_ulist: bool,
    ) -> Tuple[bool, Optional[str], Optional[RequeueLineInfo]]:
        POGGER.debug(
            "total=$;ws-before=$;ws_after=$;start_index=$",
            indent_level,
            ws_before_marker,
            ws_after_marker,
            position_marker.index_number,
        )
        POGGER.debug("extracted_whitespace=$=", extracted_whitespace)
        if indent_level >= 0:
            POGGER.debug("indent_level=$=", indent_level)
            POGGER.debug("ws_before_marker=$=", ws_before_marker)
            POGGER.debug("forced_container_whitespace=$=", forced_container_whitespace)
            return ListBlockCreateNewHandler.create_new_list(
                parser_state,
                position_marker,
                indent_level,
                extracted_whitespace,
                ws_before_marker,
                ws_after_marker,
                index,
                container_level_tokens,
                remaining_whitespace,
                after_marker_ws_index,
                current_container_blocks,
                container_depth,
                original_line,
                is_ulist,
                adj_ws,
                forced_container_whitespace,
            )
        return (
            False,
            adjusted_text_to_parse,
            None,
        )

    # pylint: enable=too-many-arguments, too-many-locals

    @staticmethod
    def __handle_list_block_pull_from_grab_bag(
        grab_bag: ContainerGrabBag,
    ) -> Tuple[
        str,
        str,
        BlockQuoteData,
        int,
        List[StackToken],
        int,
        bool,
        str,
    ]:
        assert (
            grab_bag.extracted_whitespace is not None
        ), "extracted_whitespace must be defined by now"
        assert (
            grab_bag.removed_chars_at_start_of_line is not None
        ), "Characters must have been removed by this point."
        assert grab_bag.adj_ws is not None, "adj_ws must be defined by now"

        extracted_whitespace = grab_bag.extracted_whitespace
        removed_chars_at_start: int = grab_bag.removed_chars_at_start_of_line
        adj_ws = grab_bag.adj_ws

        block_quote_data: BlockQuoteData = grab_bag.block_quote_data

        current_container_blocks: List[StackToken] = grab_bag.current_container_blocks
        container_depth: int = grab_bag.container_depth

        indent_already_processed: bool = grab_bag.was_indent_already_processed
        original_line = grab_bag.original_line

        return (
            extracted_whitespace,
            adj_ws,
            block_quote_data,
            removed_chars_at_start,
            current_container_blocks,
            container_depth,
            indent_already_processed,
            original_line,
        )

    # pylint: disable=too-many-locals
    @staticmethod
    def handle_list_block(
        is_ulist: bool,
        parser_state: ParserState,
        position_marker: PositionMarker,
        grab_bag: ContainerGrabBag,
    ) -> Tuple[bool, int, Optional[str], List[MarkdownToken]]:
        """
        Handle the processing of a ulist block.
        """
        (
            extracted_whitespace,
            adj_ws,
            block_quote_data,
            removed_chars_at_start,
            current_container_blocks,
            container_depth,
            indent_already_processed,
            original_line,
        ) = ListBlockProcessor.__handle_list_block_pull_from_grab_bag(grab_bag)

        (
            did_process,
            requeue_line_info,
            old_extracted_whitespace,
            old_indent_already_processed,
        ) = (
            False,
            None,
            extracted_whitespace,
            indent_already_processed,
        )
        adjusted_text_to_parse: Optional[str] = position_marker.text_to_parse
        container_level_tokens: List[MarkdownToken] = []

        (
            started_ulist,
            end_of_ulist_start_index,
            index,
            number_of_digits,
            indent_already_used,
            forced_container_whitespace,
            extracted_whitespace,
            adj_ws,
            indent_already_processed,
        ) = ListBlockProcessor.__handle_list_block_init(
            parser_state,
            position_marker,
            extracted_whitespace,
            adj_ws,
            is_ulist,
            container_depth,
            removed_chars_at_start,
            indent_already_processed,
        )
        POGGER.debug("clt>>list-start=$", started_ulist)
        if started_ulist:
            POGGER.debug("clt>>ulist-start")
            removed_chars_at_start = indent_already_used
            assert (
                index is not None and number_of_digits is not None
            ), "If started_ulist, both index and number_of_digits must be valid."
            (
                indent_level,
                remaining_whitespace,
                ws_after_marker,
                after_marker_ws_index,
                ws_before_marker,
                container_level_tokens,
                block_quote_data,
            ) = ListBlockPreListHelper.pre_list(
                parser_state,
                position_marker.text_to_parse,
                index,
                extracted_whitespace,
                number_of_digits,
                block_quote_data,
                adj_ws,
                position_marker,
                container_depth,
            )

            (
                did_process,
                adjusted_text_to_parse,
                requeue_line_info,
            ) = ListBlockProcessor.__create_new_list_if_indented(
                parser_state,
                position_marker,
                indent_level,
                ws_before_marker,
                ws_after_marker,
                extracted_whitespace,
                forced_container_whitespace,
                adj_ws,
                adjusted_text_to_parse,
                index,
                container_level_tokens,
                remaining_whitespace,
                after_marker_ws_index,
                current_container_blocks,
                container_depth,
                original_line,
                is_ulist,
            )
        else:
            extracted_whitespace = old_extracted_whitespace
            indent_already_processed = old_indent_already_processed

        ListBlockProcessor.__handle_list_block_push_to_grab_bag(
            grab_bag,
            removed_chars_at_start,
            block_quote_data,
            requeue_line_info,
            indent_already_processed,
            extracted_whitespace,
        )

        return (
            did_process,
            end_of_ulist_start_index,
            adjusted_text_to_parse,
            container_level_tokens,
        )

    # pylint: enable=too-many-locals

    # pylint: disable=too-many-arguments
    @staticmethod
    def __handle_list_block_push_to_grab_bag(
        grab_bag: ContainerGrabBag,
        removed_chars_at_start: int,
        block_quote_data: BlockQuoteData,
        requeue_line_info: Optional[RequeueLineInfo],
        indent_already_processed: bool,
        extracted_whitespace: Optional[str],
    ) -> None:
        grab_bag.removed_chars_at_start_of_line = removed_chars_at_start
        grab_bag.block_quote_data = block_quote_data
        grab_bag.requeue_line_info = requeue_line_info
        grab_bag.was_indent_already_processed = indent_already_processed
        grab_bag.extracted_whitespace = extracted_whitespace

    # pylint: enable=too-many-arguments
    # pylint: disable=too-many-arguments
    @staticmethod
    def __list_in_process_update_containers(
        parser_state: ParserState,
        ind: int,
        used_indent: Optional[str],
        was_paragraph_continuation: bool,
        start_index: int,
        grab_bag: ContainerGrabBag,
    ) -> None:
        stack_index = parser_state.find_last_list_block_on_stack()
        if stack_index > 0:
            list_token = cast(
                ListStartMarkdownToken,
                parser_state.token_stack[stack_index].matching_markdown_token,
            )
            POGGER.debug(
                "lip>>last_block_token>>$",
                list_token,
            )

        POGGER.debug(">>used_indent>>$<<", used_indent)
        POGGER.debug(">>was_paragraph_continuation>>$<<", was_paragraph_continuation)
        if used_indent is not None:
            list_token = cast(
                ListStartMarkdownToken,
                parser_state.token_stack[ind].matching_markdown_token,
            )
            POGGER.debug(
                "__list_in_process_update_containers>>list_token>>$", list_token
            )
            list_token.add_leading_spaces(used_indent)
            POGGER.debug(
                "__list_in_process_update_containers>>list_token>>$", list_token
            )
        else:
            stack_index = parser_state.find_last_list_block_on_stack()
            need_to_add_leading_spaces = False
            if stack_index > 0:
                assert (
                    parser_state.original_line_to_parse is not None
                ), "Original line must be defined by this point."

                last_container_index = parser_state.find_last_container_on_stack()
                consumed_text = parser_state.original_line_to_parse[:start_index]
                back_index = stack_index
                while back_index and parser_state.token_stack[back_index].is_list:
                    back_index -= 1
                need_to_add_leading_spaces = (
                    back_index <= 0
                    or not consumed_text
                    or ">" in consumed_text
                    or stack_index != last_container_index
                )
            if need_to_add_leading_spaces:
                ListBlockProcessor.__list_in_process_update_containers_part_2(
                    parser_state, stack_index, consumed_text, grab_bag
                )

    # pylint: enable=too-many-arguments

    @staticmethod
    def __list_in_process_update_containers_part_2(
        parser_state: ParserState,
        stack_index: int,
        consumed_text: str,
        grab_bag: ContainerGrabBag,
    ) -> None:
        list_stack_token = cast(ListStackToken, parser_state.token_stack[stack_index])
        list_token = cast(
            ListStartMarkdownToken,
            parser_state.token_stack[stack_index].matching_markdown_token,
        )
        if lists_new_list_token := list_stack_token.last_new_list_token:
            lists_indent_level = lists_new_list_token.indent_level
        else:
            lists_indent_level = list_token.indent_level
        assert not consumed_text or len(consumed_text) < lists_indent_level
        grab_bag.bogus = consumed_text
        POGGER.debug("__list_in_process_update_containers2>>list_token>>$", list_token)
        list_token.add_leading_spaces("")
        POGGER.debug("__list_in_process_update_containers2>>list_token>>$", list_token)

    # pylint: disable=too-many-locals
    @staticmethod
    def list_in_process(
        parser_state: ParserState,
        ind: int,
        index_indent: int,
        grab_bag: ContainerGrabBag,
    ) -> List[MarkdownToken]:
        """
        Handle the processing of a line where there is a list in process.
        """
        assert (
            grab_bag.extracted_whitespace is not None
        ), "Extracted whitespace must be defined by this point."
        assert parser_state.token_stack[
            ind
        ].is_list, "Index must point to a valid list."
        extracted_whitespace = grab_bag.extracted_whitespace
        assert (
            extracted_whitespace is not None
        ), "extracted_whitespace must be defined by now"

        line_to_parse = grab_bag.line_to_parse
        start_index = grab_bag.start_index
        list_stack_token = cast(ListStackToken, parser_state.token_stack[ind])
        before_ws_length = list_stack_token.ws_before_marker
        leading_space_length = TabHelper.calculate_length(extracted_whitespace)
        requested_list_indent = (
            list_stack_token.last_new_list_token.indent_level
            if list_stack_token.last_new_list_token
            else list_stack_token.indent_level
        )

        allow_list_continue = ListBlockProcessor.__can_list_continue(
            parser_state,
            line_to_parse,
            start_index,
            extracted_whitespace,
            leading_space_length,
        )
        POGGER.debug(
            "leading_space_length>>$>>requested_list_indent>>$>>is_in_paragraph>>$",
            leading_space_length,
            requested_list_indent,
            parser_state.token_stack[-1].is_paragraph,
        )
        used_indent = None
        if (
            was_paragraph_continuation := leading_space_length >= requested_list_indent
            and allow_list_continue
        ):
            POGGER.debug("list-in-progress: was_paragraph_continuation")
            container_level_tokens: List[MarkdownToken] = []
            (
                line_to_parse,
                used_indent,
            ) = ListBlockProcessor.__adjust_line_for_list_in_process(
                line_to_parse,
                start_index,
                extracted_whitespace,
                leading_space_length,
                requested_list_indent,
                grab_bag.original_line,
            )
        else:
            POGGER.debug("list-in-progress: not was_paragraph_continuation")
            (
                container_level_tokens,
                line_to_parse,
                used_indent,
                ind,
                requeue_line_info,
                was_paragraph_continuation,
            ) = ListBlockProcessor.__process_list_non_continue(
                parser_state,
                requested_list_indent,
                leading_space_length,
                before_ws_length,
                line_to_parse,
                start_index,
                extracted_whitespace,
                allow_list_continue,
                ind,
                grab_bag.original_line,
                index_indent,
            )
            if requeue_line_info:
                grab_bag.line_to_parse = line_to_parse
                grab_bag.indent_used_by_list = None
                grab_bag.requeue_line_info = requeue_line_info
                grab_bag.was_paragraph_continuation = False
                return []
        ListBlockProcessor.__list_in_process_update_containers(
            parser_state,
            ind,
            used_indent,
            was_paragraph_continuation,
            start_index,
            grab_bag,
        )

        grab_bag.line_to_parse = line_to_parse
        grab_bag.indent_used_by_list = used_indent
        grab_bag.requeue_line_info = None
        grab_bag.was_paragraph_continuation = was_paragraph_continuation

        return container_level_tokens

    # pylint: enable=too-many-locals

    @staticmethod
    def __can_list_continue(
        parser_state: ParserState,
        line_to_parse: str,
        start_index: int,
        extracted_whitespace: str,
        leading_space_length: int,
    ) -> bool:
        started_ulist, _, _, _ = ListBlockStartsHelper.is_ulist_start(
            parser_state, line_to_parse, start_index, extracted_whitespace, True, None
        )
        started_olist, _, _, _ = ListBlockStartsHelper.is_olist_start(
            parser_state, line_to_parse, start_index, extracted_whitespace, True, None
        )
        return (
            (not parser_state.token_document[-1].is_blank_line)
            if leading_space_length >= 4 and (started_ulist or started_olist)
            else True
        )

    # pylint: disable=too-many-arguments
    @staticmethod
    def __check_for_paragraph_break(
        parser_state: ParserState,
        line_to_parse: str,
        start_index: int,
        extracted_whitespace: str,
        original_line: str,
        index_indent: int,
    ) -> bool:
        POGGER.debug("is_theme_break>>?")
        is_theme_break, _ = ThematicLeafBlockProcessor.is_thematic_break(
            line_to_parse,
            start_index,
            extracted_whitespace,
            skip_whitespace_check=True,
        )
        POGGER.debug("is_theme_break>>$", is_theme_break)
        POGGER.debug("is_atx_heading>>?")
        is_atx_heading, _, _, _ = AtxLeafBlockProcessor.is_atx_heading(
            line_to_parse, start_index, extracted_whitespace, skip_whitespace_check=True
        )
        POGGER.debug("is_atx_heading>>$", is_atx_heading)
        POGGER.debug("is_fenced_start>>?")
        is_fenced_start, _, _, _, _ = FencedLeafBlockProcessor.is_fenced_code_block(
            parser_state,
            line_to_parse,
            start_index,
            extracted_whitespace,
            original_line,
            index_indent,
            skip_whitespace_check=True,
        )
        POGGER.debug("is_fenced_start>>$", is_fenced_start)
        POGGER.debug("is_html_start>>?")
        is_html_start, _ = HtmlHelper.is_html_block(
            line_to_parse,
            start_index,
            extracted_whitespace,
            parser_state.token_stack,
            parser_state.parse_properties,
            skip_whitespace_check=True,
        )
        POGGER.debug("is_html_start>>$", is_html_start)
        return (
            bool(is_theme_break)
            or is_atx_heading
            or is_fenced_start
            or bool(is_html_start)
        )

    # pylint: enable=too-many-arguments

    # pylint: disable=too-many-arguments,too-many-locals
    @staticmethod
    def __process_list_non_continue(
        parser_state: ParserState,
        requested_list_indent: int,
        leading_space_length: int,
        before_ws_length: int,
        line_to_parse: str,
        start_index: int,
        extracted_whitespace: str,
        allow_list_continue: bool,
        ind: int,
        original_line: str,
        index_indent: int,
    ) -> Tuple[
        List[MarkdownToken],
        str,
        Optional[str],
        int,
        Optional[RequeueLineInfo],
        bool,
    ]:
        POGGER.debug(
            "requested_list_indent>>$<<",
            requested_list_indent,
        )
        requested_list_indent -= before_ws_length

        POGGER.debug(
            "leading_space_length>>$>>adj requested_list_indent>>$>>$<<",
            leading_space_length,
            requested_list_indent,
            parser_state.token_stack[-1].is_paragraph,
        )

        was_breakable_leaf_detected = ListBlockProcessor.__check_for_paragraph_break(
            parser_state,
            line_to_parse,
            start_index,
            extracted_whitespace,
            original_line,
            index_indent,
        )

        if was_paragraph_continuation := (
            parser_state.token_stack[-1].is_paragraph
            and leading_space_length >= requested_list_indent
            and allow_list_continue
            and not was_breakable_leaf_detected
        ):
            container_level_tokens: List[MarkdownToken] = []
            POGGER.debug(
                "1>>line_to_parse>>$>>",
                line_to_parse,
            )
            (
                line_to_parse,
                used_indent,
            ) = ListBlockProcessor.__adjust_line_for_list_in_process(
                line_to_parse,
                start_index,
                extracted_whitespace,
                leading_space_length,
                requested_list_indent + before_ws_length,
                original_line,
            )
            was_paragraph_continuation = used_indent is None
            POGGER.debug(
                ">>line_to_parse>>$>>",
                line_to_parse,
            )
            POGGER.debug(">>used_indent>>$>>", used_indent)
        else:
            POGGER.debug(
                "2>>line_to_parse>>$>>",
                line_to_parse,
            )
            (
                container_level_tokens,
                requeue_line_info,
            ) = ListBlockProcessor.__check_for_list_closures(
                parser_state,
                line_to_parse,
                start_index,
                extracted_whitespace,
                ind,
                leading_space_length,
                original_line,
                index_indent,
            )
            POGGER.debug(
                "2>>requeue_line_info>>$>>",
                requeue_line_info,
            )
            if requeue_line_info:
                return [], line_to_parse, None, -1, requeue_line_info, False

            (
                line_to_parse,
                used_indent,
                ind,
            ) = ListBlockProcessor.__adjust_for_nested_list(
                parser_state,
                container_level_tokens,
                ind,
                line_to_parse,
                extracted_whitespace,
                start_index,
                before_ws_length,
                leading_space_length,
                original_line,
            )

        # POGGER.debug(">>line_to_parse>>$>>",line_to_parse)
        return (
            container_level_tokens,
            line_to_parse,
            used_indent,
            ind,
            None,
            was_paragraph_continuation,
        )

    # pylint: enable=too-many-arguments,too-many-locals

    # pylint: disable=too-many-arguments
    @staticmethod
    def __adjust_for_nested_list(
        parser_state: ParserState,
        container_level_tokens: List[MarkdownToken],
        ind: int,
        line_to_parse: str,
        extracted_whitespace: str,
        start_index: int,
        before_ws_length: int,
        leading_space_length: int,
        original_line: str,
    ) -> Tuple[str, Optional[str], int]:
        POGGER.debug(
            "2>>__check_for_list_closures>>$>>",
            container_level_tokens,
        )
        POGGER.debug(
            "2>>parser_state.token_stack>>$>>",
            parser_state.token_stack,
        )
        POGGER.debug(
            "2>>ind>>$>>",
            ind,
        )

        found_owning_list: Optional[ListStackToken] = None
        if container_level_tokens:
            (
                did_find,
                last_list_index,
            ) = LeafBlockProcessorParagraph.check_for_list_in_process(parser_state)
            POGGER.debug(
                "2>>did_find>>$>>$>>",
                did_find,
                last_list_index,
            )
            if did_find:
                ind = last_list_index
                found_owning_list = cast(ListStackToken, parser_state.token_stack[ind])
        else:
            assert parser_state.token_stack[ind].is_list, "Token must be a list token."
            found_owning_list = cast(ListStackToken, parser_state.token_stack[ind])

        if found_owning_list:
            POGGER.debug(">>in list>>")
            requested_list_indent = (
                (found_owning_list.last_new_list_token.indent_level)
                if found_owning_list.last_new_list_token
                else found_owning_list.indent_level
            )
            POGGER.debug(">>line_to_parse>>$>>", line_to_parse)
            POGGER.debug(">>extracted_whitespace>>$<<", extracted_whitespace)
            POGGER.debug(">>start_index>>$", start_index)
            POGGER.debug(">>requested_list_indent>>$", requested_list_indent)
            POGGER.debug(">>before_ws_length>>$", before_ws_length)
            (
                line_to_parse,
                used_indent,
            ) = ListBlockProcessor.__adjust_line_for_list_in_process(
                line_to_parse,
                start_index,
                extracted_whitespace,
                leading_space_length,
                requested_list_indent,
                original_line,
            )
            POGGER.debug(">>line_to_parse>>$", line_to_parse)
            POGGER.debug(">>used_indent>>$<<", used_indent)
        else:
            used_indent = None
        return line_to_parse, used_indent, ind

    # pylint: enable=too-many-arguments

    @staticmethod
    def __adjust_line_for_list_in_process_with_tab(
        original_line: str, remaining_indent: int, removed_whitespace: str
    ) -> str:
        _, ex_ws = ParserHelper.extract_spaces_verified(original_line, 0)
        POGGER.debug("ex_ws($)", ex_ws)
        if "\t" in ex_ws:
            detabified_ws = TabHelper.detabify_string(ex_ws, 0)
            POGGER.debug("detabified_ws($)", detabified_ws)
            assert (
                len(detabified_ws) >= remaining_indent
            ), "Length must be greater than the indent."

            sub_ws = None
            i = -1
            do_loop = True
            # for i in range(len(ex_ws)):
            while do_loop:
                i += 1
                assert i < len(ex_ws), "Index must be within the string."
                sub_ws = ex_ws[: i + 1]
                POGGER.debug("sub_ws($)", sub_ws)
                detabified_ws = TabHelper.detabify_string(sub_ws, 0)
                POGGER.debug("detabified_ws($)", detabified_ws)
                POGGER.debug(
                    "detabified_ws($),removed_whitespace($)",
                    len(detabified_ws),
                    len(removed_whitespace),
                )
                if len(detabified_ws) >= len(removed_whitespace):
                    do_loop = False
            if len(detabified_ws) <= len(removed_whitespace):
                assert sub_ws is not None, "Subws must have been defined."
                removed_whitespace = sub_ws
        return removed_whitespace

    # pylint: disable=too-many-arguments
    @staticmethod
    def __adjust_line_for_list_in_process(
        line_to_parse: str,
        start_index: int,
        leading_space: str,
        leading_space_length: int,
        requested_list_indent: int,
        original_line: str,
    ) -> Tuple[str, Optional[str]]:
        """
        Alter the current line to better represent the current level of lists.
        """
        remaining_indent = leading_space_length - requested_list_indent
        POGGER.debug("original_line($)", original_line)
        POGGER.debug(
            "enough ws to continue; line($),start_index($),leading_space($)",
            line_to_parse,
            start_index,
            leading_space,
        )
        if remaining_indent < 0:
            padded_spaces = ""
            start_index = 0
            removed_whitespace = None
        else:
            POGGER.debug("requested_list_indent($)", requested_list_indent)
            POGGER.debug("leading_space($)", leading_space)
            padded_spaces = ParserHelper.repeat_string(
                ParserHelper.space_character, remaining_indent
            )
            removed_whitespace = leading_space[:requested_list_indent]
            if "\t" in original_line:
                removed_whitespace = (
                    ListBlockProcessor.__adjust_line_for_list_in_process_with_tab(
                        original_line, remaining_indent, removed_whitespace
                    )
                )
        POGGER.debug("padded_spaces($)", padded_spaces)
        POGGER.debug("line_to_parse[start_index:]($)", line_to_parse[start_index:])
        POGGER.debug("removed_whitespace($)", removed_whitespace)
        return (
            f"{padded_spaces}{line_to_parse[start_index:]}",
            removed_whitespace,
        )

    # pylint: enable=too-many-arguments

    # pylint: disable=too-many-arguments
    @staticmethod
    def __check_for_list_closures(
        parser_state: ParserState,
        line_to_parse: str,
        start_index: int,
        extracted_whitespace: str,
        ind: int,
        leading_space_length: int,
        original_line: str,
        index_indent: int,
    ) -> Tuple[List[MarkdownToken], Optional[RequeueLineInfo]]:
        """
        Check to see if the list in progress and the level of lists shown require
        the closing of some of the sublists.
        """
        POGGER.debug("ws(naa)>>line_to_parse>>$<<", line_to_parse)
        POGGER.debug("ws(naa)>>stack>>$", parser_state.token_stack)
        POGGER.debug("ws(naa)>>tokens>>$", parser_state.token_document)

        is_leaf_block_start = LeafBlockProcessor.is_paragraph_ending_leaf_block_start(
            parser_state,
            line_to_parse,
            start_index,
            extracted_whitespace,
            original_line,
            index_indent,
        )
        if not parser_state.token_stack[-1].is_paragraph or is_leaf_block_start:
            POGGER.debug("ws (normal and adjusted) not enough to continue")

            POGGER.debug("lsl $", leading_space_length)
            POGGER.debug("lsl $", parser_state.token_stack[ind])
            search_index = ind
            POGGER.debug(
                "lsl $>$",
                search_index,
                parser_state.token_stack[search_index - 1],
            )
            while parser_state.token_stack[search_index - 1].is_list:
                list_token = cast(
                    ListStackToken, parser_state.token_stack[search_index - 1]
                )
                if list_token.indent_level <= leading_space_length:
                    break
                search_index -= 1

            POGGER.debug("lsl $", parser_state.token_stack[search_index])

            (
                container_level_tokens,
                requeue_line_info,
            ) = parser_state.close_open_blocks_fn(
                parser_state,
                until_this_index=search_index,
                include_lists=True,
                caller_can_handle_requeue=True,
                requeue_reset=True,
            )
        else:
            container_level_tokens, requeue_line_info = [], None

        return container_level_tokens, requeue_line_info

    # pylint: enable=too-many-arguments



"""
Module to provide for a debug extension.
"""

from pymarkdown.extension_manager.extension_impl import ExtensionDetails
from pymarkdown.extension_manager.extension_manager_constants import (
    ExtensionManagerConstants,
)
from pymarkdown.extension_manager.parser_extension import ParserExtension
from pymarkdown.my_application_properties_facade import MyApplicationPropertiesFacade


class ExceptionTestException(Exception):
    """
    Custom exception for the DebugExtension class.
    """

    def __init__(self, message: str) -> None:
        self.message = message
        super().__init__(self.message)


class DebugExtension(ParserExtension):
    """
    Extension to provide for a debug extension.
    """

    def get_identifier(self) -> str:
        """
        Get the identifier associated with this extension.
        """
        return "debug-extension"

    def get_details(self) -> ExtensionDetails:
        """
        Get the details for the extension.
        """
        return ExtensionDetails(
            extension_id=self.get_identifier(),
            extension_name="Debug Extension",
            extension_description="Allows testing through debug.",
            extension_enabled_by_default=False,
            extension_version=ExtensionManagerConstants.EXTENSION_VERSION_NOT_IMPLEMENTED,
            extension_interface_version=ExtensionManagerConstants.EXTENSION_INTERFACE_VERSION_BASIC,
            extension_url=None,
            extension_configuration=None,
        )

    def apply_configuration(
        self, extension_specific_facade: MyApplicationPropertiesFacade
    ) -> None:
        """
        Apply any configuration required by the extension.
        """
        debug_mode = extension_specific_facade.get_integer_property(
            "debug_mode", default_value=0
        )
        if debug_mode == 1:
            raise ExceptionTestException("blah")
        if debug_mode == 2:
            raise ValueError("blah")



"""
Module to provide for the creation of a new link.
"""

import logging
from typing import List, Optional, Tuple, cast

from typing_extensions import Protocol

from pymarkdown.container_blocks.parse_block_pass_properties import (
    ParseBlockPassProperties,
)
from pymarkdown.general.parser_helper import ParserHelper
from pymarkdown.general.parser_logger import ParserLogger
from pymarkdown.inline.inline_helper import InlineHelper
from pymarkdown.links.link_helper_properties import LinkHelperProperties
from pymarkdown.tokens.email_autolink_markdown_token import EmailAutolinkMarkdownToken
from pymarkdown.tokens.image_start_markdown_token import ImageStartMarkdownToken
from pymarkdown.tokens.inline_code_span_markdown_token import (
    InlineCodeSpanMarkdownToken,
)
from pymarkdown.tokens.link_start_markdown_token import LinkStartMarkdownToken
from pymarkdown.tokens.markdown_token import MarkdownToken
from pymarkdown.tokens.raw_html_markdown_token import RawHtmlMarkdownToken
from pymarkdown.tokens.text_markdown_token import TextMarkdownToken

POGGER = ParserLogger(logging.getLogger(__name__))


# pylint: disable=too-few-public-methods
class ProcessSimpleInlineProtocol(Protocol):
    """
    Protocol to allow for simple processing of inline elements.
    """

    def __call__(  # noqa: E704
        self,
        parser_properties: ParseBlockPassProperties,
        source_text: str,
    ) -> str: ...  # pragma: no cover


# pylint: enable=too-few-public-methods


# pylint: disable=too-few-public-methods
class LinkCreateHelper:
    """
    Class to provide for the creation of a new link.
    """

    __link_start_sequence = "["
    image_start_sequence = "!["

    # pylint: disable=too-many-arguments
    @staticmethod
    def create_link_token(
        parser_properties: ParseBlockPassProperties,
        start_text: str,
        text_from_blocks: str,
        text_from_blocks_raw: str,
        inline_blocks: List[MarkdownToken],
        ind: int,
        remaining_line: str,
        current_string_unresolved: str,
        process_inlines_fn: ProcessSimpleInlineProtocol,
        lhp: LinkHelperProperties,
        update_index: int,
    ) -> Tuple[int, Optional[MarkdownToken], bool]:
        """
        Create the right type of link token.
        """
        (
            text_from_blocks_raw,
            line_number,
            column_number,
        ) = (
            ParserHelper.resolve_backspaces_from_text(text_from_blocks_raw),
            inline_blocks[ind].line_number,
            inline_blocks[ind].column_number,
        )

        # POGGER.debug("<<<<<<<start_text<<<<<<<$<<", start_text)
        # POGGER.debug(">>inline_link>>$>>", inline_link)
        # POGGER.debug(">>pre_inline_link>>$>>", pre_inline_link)
        # POGGER.debug(">>inline_title>>$>>", inline_title)
        # POGGER.debug(">>pre_inline_title>>$>>", pre_inline_title)
        # POGGER.debug(
        #     ">>text_from_blocks>>$>>",
        #     text_from_blocks,
        # )
        _ = text_from_blocks
        if lhp.pre_inline_link == lhp.inline_link:
            lhp.pre_inline_link = ""
        if lhp.pre_inline_title == lhp.inline_title:
            lhp.pre_inline_title = ""
        # POGGER.debug(">>pre_inline_link>>$>>", pre_inline_link)

        # POGGER.debug(">>text_from_blocks_raw>>$>>", text_from_blocks_raw)
        # POGGER.debug(">>inline_blocks[ind]>>$>>", inline_blocks[ind])

        if start_text == LinkCreateHelper.__link_start_sequence:
            consume_rest_of_line = False
            token_to_append = LinkCreateHelper.__add_link_start_token(
                inline_blocks,
                ind,
                text_from_blocks_raw,
                line_number,
                column_number,
                lhp,
            )
        else:
            token_to_append = None
            (
                consume_rest_of_line,
                text_from_blocks_raw,
            ) = LinkCreateHelper.__add_image_token(
                parser_properties,
                start_text,
                inline_blocks,
                ind,
                remaining_line,
                text_from_blocks_raw,
                process_inlines_fn,
                line_number,
                column_number,
                current_string_unresolved,
                lhp,
            )
        return update_index, token_to_append, consume_rest_of_line

    # pylint: enable=too-many-arguments

    # pylint: disable=too-many-arguments
    @staticmethod
    def __add_link_start_token(
        inline_blocks: List[MarkdownToken],
        ind: int,
        text_from_blocks_raw: str,
        line_number: int,
        column_number: int,
        lhp: LinkHelperProperties,
    ) -> MarkdownToken:
        inline_blocks[ind] = LinkStartMarkdownToken(
            text_from_blocks_raw, line_number, column_number, lhp
        )
        return inline_blocks[ind].generate_close_markdown_token_from_markdown_token(
            "", ""
        )

    # pylint: enable=too-many-arguments

    # pylint: disable=too-many-arguments
    @staticmethod
    def __add_image_token(
        parser_properties: ParseBlockPassProperties,
        start_text: str,
        inline_blocks: List[MarkdownToken],
        ind: int,
        remaining_line: str,
        text_from_blocks_raw: str,
        process_inlines_fn: ProcessSimpleInlineProtocol,
        line_number: int,
        column_number: int,
        current_string_unresolved: str,
        lhp: LinkHelperProperties,
    ) -> Tuple[bool, str]:
        assert (
            start_text == LinkCreateHelper.image_start_sequence
        ), "This should be the image start sequence."
        POGGER.debug("\n>>__consume_text_for_image_alt_text>>$>>", inline_blocks)
        POGGER.debug("\n>>__consume_text_for_image_alt_text>>$>>", ind)
        POGGER.debug("\n>>__consume_text_for_image_alt_text>>$>>", remaining_line)
        (
            image_alt_text,
            text_from_blocks_raw,
        ) = LinkCreateHelper.__consume_text_for_image_alt_text(
            parser_properties,
            inline_blocks,
            ind,
            remaining_line,
            text_from_blocks_raw,
            process_inlines_fn,
        )
        POGGER.debug("\n>>__consume_text_for_image_alt_text>>$>>", image_alt_text)

        inline_blocks[ind] = ImageStartMarkdownToken(
            image_alt_text,
            text_from_blocks_raw,
            line_number,
            column_number,
            lhp,
        )
        POGGER.debug("\n>>Image>>$", inline_blocks)
        POGGER.debug(">>start_text>>$<<", start_text)
        POGGER.debug(">>remaining_line>>$<<", remaining_line)
        POGGER.debug(">>current_string_unresolved>>$<<", current_string_unresolved)
        return True, text_from_blocks_raw

    # pylint: enable=too-many-arguments

    # pylint: disable=too-many-arguments
    @staticmethod
    def __consume_text_for_image_alt_text(
        parser_properties: ParseBlockPassProperties,
        inline_blocks: List[MarkdownToken],
        ind: int,
        remaining_line: str,
        text_from_blocks_raw: str,
        process_inlines_fn: ProcessSimpleInlineProtocol,
    ) -> Tuple[str, str]:
        """
        Consume text from the inline blocks to use as part of the image's alt text.
        """

        inline_blocks_size, ind_plus_one = len(inline_blocks), ind + 1

        POGGER.debug("inline_blocks_size>>$<<", inline_blocks_size)
        POGGER.debug("ind>>$<<", ind)
        POGGER.debug(">>$<<", inline_blocks[ind_plus_one:])
        if inline_blocks_size > (ind_plus_one):
            alt_text_parts: List[str] = []

            while inline_blocks_size > ind_plus_one:
                LinkCreateHelper.__handle_next_alt_text(
                    inline_blocks, ind_plus_one, alt_text_parts
                )

                del inline_blocks[ind_plus_one]
                inline_blocks_size -= 1

            alt_text_parts.append(remaining_line)
            image_alt_text = "".join(alt_text_parts)
            POGGER.debug(">>after>>$>>", image_alt_text)
        else:
            POGGER.debug(">>composing>>$>>", text_from_blocks_raw)
            image_alt_text = process_inlines_fn(parser_properties, text_from_blocks_raw)
            image_alt_text = ParserHelper.resolve_all_from_text(image_alt_text)
            image_alt_text = InlineHelper.append_text(
                "", image_alt_text, add_text_signature=False
            )
            POGGER.debug(">>composed>>$>>", image_alt_text)

        POGGER.debug(">>image_alt_text>>$>>", image_alt_text)
        POGGER.debug(">>text_from_blocks_raw>>$>>", text_from_blocks_raw)
        return image_alt_text, text_from_blocks_raw

    # pylint: enable=too-many-arguments

    @staticmethod
    def __handle_next_alt_text_special_text(
        next_token: MarkdownToken, alt_text_parts: List[str]
    ) -> None:
        text_token = cast(TextMarkdownToken, (next_token))
        if text_token.token_text == "]":
            alt_text_parts.append(text_token.token_text)

    @staticmethod
    def __handle_next_alt_text_normal_text(
        next_token: MarkdownToken, alt_text_parts: List[str]
    ) -> None:
        text_token = cast(TextMarkdownToken, next_token)
        alt_text_parts.append(ParserHelper.resolve_all_from_text(text_token.token_text))

    @staticmethod
    def __handle_next_alt_text_raw_html(
        next_token: MarkdownToken, alt_text_parts: List[str]
    ) -> None:
        rawhtml_token = cast(RawHtmlMarkdownToken, next_token)
        alt_text_parts.extend(["<", rawhtml_token.raw_tag, ">"])

    @staticmethod
    def __handle_next_alt_text_code_span(
        next_token: MarkdownToken, alt_text_parts: List[str]
    ) -> None:
        codespan_token = cast(InlineCodeSpanMarkdownToken, next_token)
        alt_text_parts.append(
            ParserHelper.resolve_all_from_text(codespan_token.span_text)
        )

    @staticmethod
    def __handle_next_alt_text_auto_link(
        next_token: MarkdownToken, alt_text_parts: List[str]
    ) -> None:
        autolink_token = cast(EmailAutolinkMarkdownToken, next_token)
        alt_text_parts.append(
            ParserHelper.resolve_all_from_text(autolink_token.autolink_text)
        )

    @staticmethod
    def __handle_next_alt_text_hard_break(
        next_token: MarkdownToken, alt_text_parts: List[str]
    ) -> None:
        _ = next_token
        alt_text_parts.append(ParserHelper.newline_character)

    @staticmethod
    def __handle_next_alt_text_else(
        next_token: MarkdownToken, alt_text_parts: List[str]
    ) -> None:
        assert next_token.is_inline_image, "This should be an inline image token."
        image_token = cast(ImageStartMarkdownToken, next_token)
        alt_text_parts.append(image_token.image_alt_text)

    @staticmethod
    def __handle_next_alt_text(
        inline_blocks: List[MarkdownToken], ind_plus_one: int, alt_text_parts: List[str]
    ) -> None:
        if inline_blocks[ind_plus_one].is_special_text:
            LinkCreateHelper.__handle_next_alt_text_special_text(
                inline_blocks[ind_plus_one], alt_text_parts
            )
        elif inline_blocks[ind_plus_one].is_text:
            LinkCreateHelper.__handle_next_alt_text_normal_text(
                inline_blocks[ind_plus_one], alt_text_parts
            )
        elif inline_blocks[ind_plus_one].is_inline_raw_html:
            LinkCreateHelper.__handle_next_alt_text_raw_html(
                inline_blocks[ind_plus_one], alt_text_parts
            )
        elif inline_blocks[ind_plus_one].is_inline_code_span:
            LinkCreateHelper.__handle_next_alt_text_code_span(
                inline_blocks[ind_plus_one], alt_text_parts
            )
        elif inline_blocks[ind_plus_one].is_inline_autolink:
            LinkCreateHelper.__handle_next_alt_text_auto_link(
                inline_blocks[ind_plus_one], alt_text_parts
            )
        elif (
            inline_blocks[ind_plus_one].is_inline_link
            or inline_blocks[ind_plus_one].is_inline_link_end
            or inline_blocks[ind_plus_one].is_inline_emphasis
            or inline_blocks[ind_plus_one].is_inline_emphasis_end
        ):
            pass
        elif inline_blocks[ind_plus_one].is_inline_hard_break:
            LinkCreateHelper.__handle_next_alt_text_hard_break(
                inline_blocks[ind_plus_one], alt_text_parts
            )
        else:
            LinkCreateHelper.__handle_next_alt_text_else(
                inline_blocks[ind_plus_one], alt_text_parts
            )


# pylint: enable=too-few-public-methods



"""
Module containing extra information for a link reference.
"""

from dataclasses import dataclass
from typing import Optional


@dataclass(frozen=True)
class LinkReferenceInfo:
    """
    Class containing extra information for a link reference.
    """

    collected_destination: Optional[str]
    line_destination_whitespace: Optional[str]
    inline_raw_link: Optional[str]
    line_title_whitespace: Optional[str]
    inline_raw_title: Optional[str]
    end_whitespace: Optional[str]
